import os
import sys
import argparse
import re

from datetime import timedelta, datetime


try:
    from schrodinger.application.desmond.packages import topo, traj, traj_util
    from schrodinger.structutils import analyze
except ImportError:
    print("Schrodinger Python API is required")
    sys.exit(1)


def batch_report() -> None:

    # requires Schrodinger Python API
    subprocess.run([schrodinger_run, script_path.joinpath('batch-desmond-report.py')])

    run = os.environ["SCHRODINGER"]+"/run"

    excluded_residues = ["POPC", "SPC", "T3P", "TIP3P", "T4P", "TIP4P", "NA", "K", "CL"]

    parser = argparse.ArgumentParser(description="batch gdesmond md jobs",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-p', '--protein', dest="protein", default="", help="protein ASL")
    parser.add_argument('-l', '--ligand', dest="ligand", default="none", help="ligand ASL")
    parser.add_argument('-s', '--slice-trj', dest="slice", default="", help="slice trajectory (start:end:step)")
    parser.add_argument('-w', '--overwrite', dest="overwrite", default=False, action="store_true", help="overwrite output(s)")
    parser.add_argument('-j', '--jobfile', dest="job_file", default="desmond_report_1.sh", help="job filename")
    parser.add_argument('cms', nargs="+", help="...-out.cms file")
    args = parser.parse_args()

    job_file = args.job_file
    while os.path.exists(job_file):
        splited = job_file.replace(".sh","").split("_")
        splited[-1] = str(int(splited[-1]) + 1)
        job_file = "_".join(splited) + ".sh"

    if args.slice:
        slice_trj = f'-slice-trj {args.slice}'
    else:
        slice_trj = ''

    # ex. md_1/r00/desmond_md_job_00_md2_m_naloxone-out.cms
    # mandatory input
    if not args.cms :
        parser.print_help()
        sys.exit(0)

    # provide options if protein and ligand ASL are not given
    if not (args.protein and args.ligand):
        ASL_choice = []
        protein_asl_index = None
        ligand_asl_index = None

        # read .cms and its associated trajectory file: <jobname>-out.cms, <jobname>_trj
        input_cms = args.cms[0]

        try:
            (msys_model, cms_model, trj) = traj_util.read_cms_and_traj(input_cms)
        except:
            (msys_model, cms_model) = topo.read_cms(input_cms)
            try:
                trj_path = topo.find_traj_path(cms_model)  # or
            except:
                try:
                    trj_path = topo.find_traj_path_from_cms_path(input_cms)
                except:
                    trj_path = input_cms.replace("-out.cms","_trj")
            trj = traj.read_traj(trj_path)
            
        print()
        print("Title:", cms_model.title)
        print("Excluded residues:", ",".join(excluded_residues))
        print()
        
        # chain based ASL choices
        for chain in cms_model.chain:
            # r and s for the first and last residues
            for r in chain.residue:
                break
            res = [ s for s in chain.residue if s.molecule_number == r.molecule_number ]
            s = res[-1]
            if (r.pdbres.strip() in excluded_residues) or (s.pdbres.strip() in excluded_residues):
                continue
            if chain.name.strip() :
                ASL_choice.append((
                    f"mol. {r.molecule_number} and chain. {chain.name}",
                    f"{len(res):6d} residues ({r.pdbres}{r.resnum}...{s.pdbres}{s.resnum})",
                    ))
            else: # if chain.name is blank
                ASL_choice.append((
                    f"mol. {r.molecule_number}",
                    f"{len(res):6d} residues ({r.pdbres}{r.resnum}...{s.pdbres}{s.resnum})",
                    ))

        # molecule based ASL choices
        for molecule in cms_model.molecule:
            r = None # first residue
            s = None # last residue
            for s in molecule.residue:
                if not r:
                    r = s
                pass
            if r.pdbres.strip() in excluded_residues :
                continue
            ASL_choice.append((
                f"mol. {molecule.number}",
                f"{len(molecule.residue):6d} residues ({r.pdbres}{r.resnum}...{s.pdbres}{s.resnum})",
                ))

        # sort by ASL expression text
        ASL_choice = sorted(ASL_choice)
        ASL_choice.append(("none", ""))

        # show choices
        # default
        protein_asl_index = 0
        ligand_asl_index = len(ASL_choice)-1
        for idx, (asl, info) in enumerate(ASL_choice):
            print(f"[{idx}] {asl:20s} {info}")
        print()
        
        # set protein ASL
        args.protein = ASL_choice[protein_asl_index][0] # default
        ret = input(f'Enter protein of interest (number) [{protein_asl_index}] or ASL: ')
        ret = ret.strip()
        if ret:
            if re.match('^[0-9]', ret):
                protein_asl_index = int(ret)
                args.protein = ASL_choice[protein_asl_index][0]
            else:
                args.protein = ret
        # check protein ASL
        protein_atoms = analyze.evaluate_asl(cms_model, args.protein)
        print(f"  -prot {args.protein}  [{len(protein_atoms)} atoms]")

        # set ligand ASL
        args.ligand = ASL_choice[ligand_asl_index][0] # default
        ret = input(f'Enter ligand  of interest (number) [{ligand_asl_index}] or ASL: ')
        ret = ret.strip()
        if ret:
            if re.match('^[0-9]', ret):
                ligand_asl_index = int(ret)
                args.ligand = ASL_choice[ligand_asl_index][0]
            else:
                args.ligand = ret   
        # check ligand ASL
        if not args.ligand.startswith("none"):  
            ligand_atoms = analyze.evaluate_asl(cms_model, args.ligand)
            print(f"  -lig {args.ligand}  [{len(ligand_atoms)} atoms]")

    total_cms_files = len(args.cms)
    count_cms_files = 0
    with open(job_file,"w") as job:
        for cms_file in args.cms :
            cms_base = os.path.basename(cms_file).replace("-out.cms","")
            cms_prefix = cms_file[:-8]
            trj_dir  = f"{cms_prefix}_trj"
            name = cms_base.replace("desmond_md_job_", "")
            count_cms_files += 1

            if (not os.path.exists(f'{name}-in.eaf')) or args.overwrite:
                job.write(f'{run} event_analysis.py analyze \\\n')
                job.write(f'\t{cms_file} \\\n')
                job.write(f'\t-prot "{args.protein}" \\\n')
                job.write(f'\t-lig "{args.ligand}" \\\n')
                job.write(f'\t-out {name}\n\n')

            if (not os.path.exists(f'{name}-out.eaf')) or args.overwrite:
                """
                -s START:END:STEP, -slice-trj START:END:STEP
                                Use the sliced trajectory. We use Python's slice
                                notation. START, END, and STEP should be integer
                                numbers.
                """
                job.write(f'{run} analyze_simulation.py \\\n')
                job.write(f'\t{slice_trj} {cms_file} \\\n')
                job.write(f'\t{trj_dir} \\\n')
                job.write(f'\t{name}-out.eaf \\\n')
                job.write(f'\t{name}-in.eaf\n\n')

            if (not os.path.exists(f'report_{name}.pdf')) or args.overwrite:
                job.write(f'{run} event_analysis.py report -pdf \\\n')
                job.write(f'\treport_{name}.pdf \\\n')
                job.write(f'\t{name}-out.eaf\n\n')

    os.chmod(job_file, 0o777)


import os
import sys
import argparse
import random
import re
import copy

from desmondtools import Multisim



def index_or_asl(expr:str) -> str:
    if len(expr.split()) == 1:
        return expr
    else:
        return f'"{expr}"'
    

def batch_metad():
    """Metadynamics Simulations.
    """

    SCHRODINGER = os.environ["SCHRODINGER"]
    multisim = "{}/utilities/multisim".format(SCHRODINGER)

    metad_msj = Multisim(template="desmond-metad.msj")
    metad_cfg = Multisim(template="desmond-metad.cfg")

    opt = '-HOST localhost -maxjob 1 -cpu 1 -mode umbrella -lic "DESMOND_GPGPU:16" '
    opt += '-description "metadynamics"'

    parser = argparse.ArgumentParser(description="batch gdesmond metadynamics jobs",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('--meta-height', dest="meta_height", type=float, default=0.03, help="Height of the repulsive Gaussian potential in kcal/mol (0.03)")
    parser.add_argument('--meta-interval', dest="meta_interval", type=float, default=0.09, help="Interval in ps at which Gaussian potentials are added (0.09)")
    parser.add_argument('--meta-first', dest="meta_first", type=float, default=0.0, help="Time in ps at which the Gaussian potentials are first added (0)")
    parser.add_argument('--meta-last', dest="meta_last", type=float, default=-1.0, help="Time in ps at which the Gaussian potentials are last added (simulation time)")
    parser.add_argument('--meta-kTemp', dest="meta_kTemp", type=float, default=2.4, help="Perform the well-tempered metadynamics (2.4)")
    parser.add_argument('--cv-dist', dest='cv_dist', nargs=4, action='append', default=[], help="atom1 atom2 width(0.05) wall")
    parser.add_argument('--cv-angle', dest='cv_angle', nargs=4, action='append', default=[], help="atom1 atom2 atom3 width(2.5)")
    parser.add_argument('--cv-dihedral', dest='cv_dihedral', nargs=5, action='append', default=[], help="atom1 atom2 atom3 atom4 width(5.0)")
    parser.add_argument('--cv-rgyr', dest='cv_rgyr', nargs=2, action='append', default=[], help="atom width(0.1)")
    parser.add_argument('--cv-rgyr-mass', dest='cv_rgyr_mass', nargs=2, action='append', default=[], help="atom width(0.1)")
    parser.add_argument('--cv-rmsd', dest='cv_rmsd', nargs=2, action='append', default=[], help="atom width(0.1)")
    parser.add_argument('--cv-rmsd-symm', dest='cv_rmsd_symm', nargs=2, action='append', default=[], help="atom width(0.1)")
    parser.add_argument('--cv-zdist', dest='cv_zdist', nargs=3, action='append', default=[], help="atom width wall(0.05)")
    parser.add_argument('--cv-zdist0', dest='cv_zdist0', nargs=3, action='append', default=[], help="atom width wall(0.1)")
    parser.add_argument('-g', dest="gpu_device", type=int, default=0, help="gpu device id")
    parser.add_argument('-T', dest="temperature", type=float, default=300.0, help="temperature in Kelvin")
    parser.add_argument('-t', dest="simulation_time", type=float, default=40.0, help="simulation time in ns")
    parser.add_argument('-i', dest="interval", type=float, default=40.0, help="frame interval in ps")
    parser.add_argument('-p', dest="prefix", type=str, default="m", help="directory prefix")
    parser.add_argument('-s', dest="start", type=int, default=1, help="directory start")
    parser.add_argument('-r', dest="repeat", type=int, default=1, help="number of repeats")
    parser.add_argument('-j', dest="job_file", type=str, default="desmond_metadynamics_job_1.sh", help="job filename")
    parser.add_argument('cms', nargs="+", help="desmond cms file")
    args = parser.parse_args()

    try:
        cms_files = [os.path.abspath(f) for f in args.cms]
        assert(len(cms_files) > 0)
    except:
        print(".cms file(s) not found")
        sys.exit(0)

    if not args.interval:
        args.interval = args.simulation_time # it will save 1000 frames

    job_file = args.job_file
    while os.path.exists(job_file):
        splited = job_file.replace(".sh","").split("_")
        splited[-1] = str(int(splited[-1]) + 1)
        job_file = "_".join(splited) + ".sh"


    with open("README","a") as readme, open(job_file,"w") as job:
        print("\n" + job_file + "\n")
        outdir_nums = list(range(args.start, args.start+args.repeat))
        outdirs = [f"{args.prefix}{num:02d}" for num in outdir_nums]
        cmd_echo = ""
        for argv in sys.argv:
            if cmd_echo:
                cmd_echo += " "
            if " " in argv:
                cmd_echo += f'"{argv}"'
            else:
                cmd_echo += f'{argv}'
        readme.write(f"{cmd_echo}\n\n")
        readme.write(f"GPU device              = {args.gpu_device}\n")
        readme.write(f"Temperature (K)         = {args.temperature}\n")
        readme.write(f"Simulation Time (ns)    = {args.simulation_time}\n")
        readme.write(f"Trajectory interval(ps) = {args.interval}\n")
        readme.write(f"Repeat                  = {args.repeat}\n")
        readme.write( "Directory               = %s\n" % " ".join(outdirs))
        readme.write(f"Jobfile                 = {job_file}\n\n")

        job.write(f'export CUDA_VISIBLE_DEVICES="{args.gpu_device}"\n\n')

        for i, infile in enumerate(cms_files):
            info = f"[{i+1}] {infile}"
            print(info)
            readme.write(info+"\n")
        print()
        readme.write("\n")

        for (num, outdir) in zip(outdir_nums, outdirs):
            outdir_abspath = os.path.abspath(outdir)
            job.write(f"cd {outdir_abspath}/\n\n")
            if not os.path.exists(outdir):
                os.makedirs(outdir)
            cfg_file = f"{outdir}/desmond_metadynamics_job_{num:02d}.cfg"
            msj_file = f"{outdir}/desmond_metadynamics_job_{num:02d}.msj"
            cfg_file_basename= os.path.basename(cfg_file)

            with open(cfg_file,"w") as cfg, open(msj_file,"w") as msj:
                metad_cfg.dot.randomize_velocity.seed = str(random.randint(1000,9999))
                metad_cfg.dot.time = str(args.simulation_time*1000.0)
                metad_cfg.dot.trajectory.interval = str(args.interval)
                metad_cfg.dot.temperature = f"[ [{args.temperature} 0] ]"
                metad_cfg.write(cfg)

                metad_msj.dot.simulate[-1].cfg_file = cfg_file_basename
                
                # metadynamics
                cv_list = []


                for atom1, atom2, width, wall in args.cv_dist:
                    cv = copy.deepcopy(metad_msj.dot.simulate[-1].meta.cv[0])
                    atom1 = index_or_asl(atom1)
                    atom2 = index_or_asl(atom2)
                    cv.atom = f'[{atom1} {atom2}]'
                    cv.type = 'dist'
                    cv.width = str(width)
                    if not wall.startswith('-'):
                        cv.wall = str(wall)
                    cv_list.append(cv)
                
                for atom1, atom2, atom3, width in args.cv_angle:
                    cv = copy.deepcopy(metad_msj.dot.simulate[-1].meta.cv[0])
                    atom1 = index_or_asl(atom1)
                    atom2 = index_or_asl(atom2)
                    atom3 = index_or_asl(atom3)
                    cv.atom = f'[{atom1} {atom2} {atom3}]'
                    cv.type = 'angle'
                    cv.width = str(width)
                    cv_list.append(cv)
                
                for atom1, atom2, atom3, atom4, width in args.cv_dihedral:
                    cv = copy.deepcopy(metad_msj.dot.simulate[-1].meta.cv[0])
                    atom1 = index_or_asl(atom1)
                    atom2 = index_or_asl(atom2)
                    atom3 = index_or_asl(atom3)
                    atom4 = index_or_asl(atom4)
                    cv.atom = f'[{atom1} {atom2} {atom3} {atom4}]'
                    cv.type = 'dihedral'
                    cv.width = str(width)
                    cv_list.append(cv)
                
                for atom1, width in args.cv_rmsd:
                    cv = copy.deepcopy(metad_msj.dot.simulate[-1].meta.cv[0])
                    atom1 = index_or_asl(atom1)
                    cv.atom = f'[{atom1}]'
                    cv.type = 'rmsd'
                    cv.width = str(width)
                    cv_list.append(cv)
                
                for atom1, width in args.cv_rmsd_symm:
                    cv = copy.deepcopy(metad_msj.dot.simulate[-1].meta.cv[0])
                    atom1 = index_or_asl(atom1)
                    cv.atom = f'[{atom1}]'
                    cv.type = 'rmsd_symm'
                    cv.width = str(width)
                    cv_list.append(cv)
                
                for atom1, width in args.cv_rgyr:
                    cv = copy.deepcopy(metad_msj.dot.simulate[-1].meta.cv[0])
                    atom1 = index_or_asl(atom1)
                    cv.atom = f'[{atom1}]'
                    cv.type = 'rgyr'
                    cv.width = str(width)
                    cv_list.append(cv)
                
                for atom1, width in args.cv_rgyr_mass:
                    cv = copy.deepcopy(metad_msj.dot.simulate[-1].meta.cv[0])
                    atom1 = index_or_asl(atom1)
                    cv.atom = f'[{atom1}]'
                    cv.type = 'rgyr_mass'
                    cv.width = str(width)
                    cv_list.append(cv)

                for atom1, width, wall in args.cv_zdist:
                    cv = copy.deepcopy(metad_msj.dot.simulate[-1].meta.cv[0])
                    atom1 = index_or_asl(atom1)
                    cv.atom = f'[{atom1}]'
                    cv.type = 'zdist'
                    cv.width = str(width)
                    if not wall.startswith('-'):
                        cv.wall = str(wall)
                    cv_list.append(cv)
                
                for atom1, width, wall in args.cv_zdist0:
                    cv = copy.deepcopy(metad_msj.dot.simulate[-1].meta.cv[0])
                    atom1 = index_or_asl(atom1)
                    cv.atom = f'[{atom1}]'
                    cv.type = 'zdist0'
                    cv.width = str(width)
                    if not wall.startswith('-'):
                        cv.wall = str(wall)
                    cv_list.append(cv)

                metad_msj.dot.simulate[-1].meta.cv = cv_list
                metad_msj.dot.simulate[-1].meta.height = str(args.meta_height)
                metad_msj.dot.simulate[-1].meta.interval = str(args.meta_interval)
                metad_msj.dot.simulate[-1].meta.first = str(args.meta_first)

                if args.meta_last > 0.0:
                    metad_msj.dot.simulate[-1].meta.last = str(args.meta_last)
                
                # well-tempered metadynamics
                if args.meta_kTemp > 0.0:
                    metad_msj.dot.simulate[-1].meta.kTemp = str(args.meta_kTemp)
                
                metad_msj.write(msj)
            
            # append to job file
            for infile in cms_files:
                prefix_ = re.sub(r'desmond_setup[-_]','', os.path.basename(infile))
                prefix  = prefix_.replace("-out.cms","")
                job_name = f"desmond_metadynamics_job_{num:02d}_{prefix}"
                job.write('{} -JOBNAME {} -m {} -c {} {} {} -o {} -WAIT\n\n'.format(
                    multisim, 
                    job_name,
                    os.path.basename(msj_file),
                    os.path.basename(cfg_file),
                    opt,
                    infile,
                    f"{job_name}-out.cms",
                ))

    os.chmod(job_file, 0o777)


import os
import sys
import argparse
import random
import re

from desmondtools import Multisim


def batch_min():
    SCHRODINGER = os.environ["SCHRODINGER"]
    multisim = "{}/utilities/multisim".format(SCHRODINGER)
    
    min_msj = Multisim(template="desmond-min.msj")
    min_cfg = Multisim(template="desmond-min.cfg")

    opt = '-HOST localhost -maxjob 1 -cpu 1 -mode umbrella '
    opt += '-lic "DESMOND_GPGPU:16" '
    opt += '-description "minimization"'

    parser = argparse.ArgumentParser(description="batch gdesmond minimization jobs",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-g', dest="gpu_device", type=int, default=0, 
                        help="gpu device id")
    parser.add_argument('-t', dest="simulation_time", type=float, default=100.0, 
                        help="simulation time in ps")
    parser.add_argument('-p', dest="prefix", type=str, default="r", 
                        help="directory prefix")
    parser.add_argument('-s', dest="start", type=int, default=1, 
                        help="directory start")
    parser.add_argument('-r', dest="repeat", type=int, default=1, 
                        help="number of repeats")
    parser.add_argument('-j', dest="job_file", type=str, default="desmond_min_job_1.sh", 
                        help="job filename")
    parser.add_argument('cms', nargs="+", help="desmond cms file")
    args = parser.parse_args()

    try:
        cms_files = [os.path.abspath(f) for f in args.cms]
        assert(len(cms_files) > 0)
    except:
        print(".cms file(s) not found")
        sys.exit(0)


    job_file = args.job_file
    while os.path.exists(job_file):
        splited = job_file.replace(".sh","").split("_")
        splited[-1] = str(int(splited[-1]) + 1)
        job_file = "_".join(splited) + ".sh"


    with open("README","a") as readme, open(job_file,"w") as job:
        print("\n" + job_file + "\n")
        outdir_nums = list(range(args.start, args.start+args.repeat))
        outdirs = [f"{args.prefix}{num:02d}" for num in outdir_nums]
        cmd_echo = ""
        for argv in sys.argv:
            if cmd_echo:
                cmd_echo += " "
            if " " in argv:
                cmd_echo += f'"{argv}"'
            else:
                cmd_echo += f'{argv}'
        readme.write(f"{cmd_echo}\n\n")
        readme.write(f"GPU device              = {args.gpu_device}\n")
        readme.write(f"Simulation Time (ns)    = {args.simulation_time}\n")
        readme.write(f"Repeat                  = {args.repeat}\n")
        readme.write( "Directory               = %s\n" % " ".join(outdirs))
        readme.write(f"Jobfile                 = {job_file}\n\n")

        job.write(f'export CUDA_VISIBLE_DEVICES="{args.gpu_device}"\n\n')

        for i, infile in enumerate(cms_files):
            info = f"[{i+1}] {infile}"
            print(info)
            readme.write(info+"\n")
        print()
        readme.write("\n")

        for (num, outdir) in zip(outdir_nums, outdirs):
            outdir_abspath = os.path.abspath(outdir)
            job.write(f"cd {outdir_abspath}/\n\n")
            if not os.path.exists(outdir):
                os.makedirs(outdir)
            cfg_file = f"{outdir}/desmond_min_job_{num:02d}.cfg"
            msj_file = f"{outdir}/desmond_min_job_{num:02d}.msj"
            cfg_file_basename= os.path.basename(cfg_file)

            with open(cfg_file,"w") as cfg, open(msj_file,"w") as msj:
                min_cfg.dot.randomize_velocity.seed = str(random.randint(1000,9999))
                min_cfg.dot.time = str(args.simulation_time) # (ps)
                min_cfg.write(cfg)

                min_msj.dot.simulate.cfg_file = cfg_file_basename
                min_msj.write(msj)
            
            # append to job file
            for infile in cms_files:
                prefix_ = re.sub(r'desmond_setup[-_]','', os.path.basename(infile))
                prefix  = prefix_.replace("-out.cms","")
                job_name = f"desmond_min_job_{num:02d}_{prefix}"
                job.write('{} -JOBNAME {} -m {} -c {} {} {} -o {} -WAIT\n\n'.format(
                    multisim, 
                    job_name,
                    os.path.basename(msj_file),
                    os.path.basename(cfg_file),
                    opt,
                    infile,
                    f"{job_name}-out.cms",
                ))

    os.chmod(job_file, 0o777)


import os
import sys
import argparse
import random
import re

from desmondtools import Multisim


def batch_md():
    SCHRODINGER = os.environ["SCHRODINGER"]
    multisim = "{}/utilities/multisim".format(SCHRODINGER)
    
    md_msj = Multisim(template="desmond-md.msj")
    md_cfg = Multisim(template="desmond-md.cfg")

    parser = argparse.ArgumentParser(description="batch gdesmond md jobs",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-g', dest="gpu_device", type=int, default=0, 
                        metavar="gpu_device", help="gpu device id")
    parser.add_argument('-T', dest="temp", nargs="+", type=float, default=[300.0,], 
                        metavar="temperature", help="temperature in K")
    parser.add_argument('-R', dest="ramp", nargs="+", type=float, default=[10.0,],
                        metavar="tempramp", help="heat and cool ramps in ps/K")
    parser.add_argument('-t', dest="simulation_time", nargs="+", type=float, default=[100.0,],
                        metavar="simulation_time", help="simulation time in ns")
    parser.add_argument('-i', dest="interval", type=float, default=100.0,
                        metavar="interval", help="frame interval in ps")
    parser.add_argument('--posres', dest="posres", default="res. UNK",
                        metavar="posres", help="ASL for positional restraints during prod. sim.")
    parser.add_argument('--posres-force', dest="posres_force", type=float, default=0.0,
                        metavar="posres_force", help="positional restraints force constant (kcal/mol/A**2)")
    parser.add_argument('-p', dest="prefix", default="r", help="directory prefix")
    parser.add_argument('-s', dest="start", type=int, default=1, help="directory start")
    parser.add_argument('-r', dest="repeat", type=int, default=1, help="number of repeats")
    parser.add_argument('-j', dest="job_file", default="desmond_md_job_1.sh", help="job filename")
    parser.add_argument('cms', nargs="+", help="desmond cms file")
    args = parser.parse_args()

    try:
        cms_files = [os.path.abspath(f) for f in args.cms]
        assert(len(cms_files) > 0)
    except:
        print(".cms file(s) not found")
        sys.exit(0)

    opt  = '-HOST localhost -maxjob 1 -cpu 1 -mode umbrella '
    opt += '-set stage[1].set_family.md.jlaunch_opt=["-gpu"] -lic "DESMOND_GPGPU:16"'

    job_file = args.job_file
    while os.path.exists(job_file):
        splited = job_file.replace(".sh","").split("_")
        splited[-1] = str(int(splited[-1]) + 1)
        job_file = "_".join(splited) + ".sh"

    if len(args.temp) > 1:
        try:
            assert len(args.temp) == len(args.simulation_time)
            assert len(args.temp) == (len(args.ramp) + 1)
        except:
            print("For a variable temperature simulaton, the number of temperatures and simulation times ")
            print("should match and temperature ramp(s) (ps/K) should be given between temperatures.")
            print("Please check -T, -t and -R options")
            sys.exit(0)
    else:
        args.ramp = [] # not used

    # Note: if not specified, 
    # all times are in the unit of ps and 
    # energy is in the unit of kcal/mol.

    with open("README","a") as readme, open(job_file,"w") as job:
        print(f"Job file = {job_file}")

        dirs = [ f"{args.prefix}{n:02d}" for n in range(args.start, args.start+args.repeat) ]

        t_schedule = [ [ args.temp[0], 0.0 ], ]

        if len(args.temp) > 1 :

            elapsed = 0
            prev_temp = args.temp[0]
            idx = 0

            for (temp,simulation_time) in zip(args.temp, args.simulation_time):


                deltaT = abs(temp - prev_temp)

                if deltaT > 0.001: # ramp
                    elapsed += args.ramp[idx] * deltaT # ramp: ps/K
                    t_schedule.append([temp, elapsed])
                    idx += 1

                elapsed += simulation_time * 1000. # ns -> ps
                t_schedule.append([temp, elapsed]) # ns -> ps
                prev_temp = temp
            total_simulation_time = elapsed # ps
        else:
            total_simulation_time = sum(args.simulation_time) * 1000.0 # ps

        if not args.interval:
            # args.simulation_time in ns and args.interval in ps
            # default: make 1000 frames
            args.interval = total_simulation_time / 1000.
        
        cmd_echo = ""
        for argv in sys.argv:
            if cmd_echo:
                cmd_echo += " "
            if " " in argv:
                cmd_echo += f'"{argv}"'
            else:
                cmd_echo += f'{argv}'
        readme.write(f"{cmd_echo}\n\n")
        readme.write(f"GPU device              = {args.gpu_device}\n")
        readme.write(f"Temperature (K)         = {args.temp}\n")
        readme.write(f"Temperature Ramp (ps/K) = {args.ramp}\n")
        readme.write(f"Simulation Time (ns)    = {args.simulation_time}\n")
        readme.write(f"Temperature schedule    = {str(t_schedule)}\n")
        readme.write(f"Total Sim. Time (ns)    = {total_simulation_time/1000.0}\n")
        readme.write(f"Trajectory interval(ps) = {args.interval}\n")
        readme.write(f"Repeat                  = {args.repeat}\n")
        readme.write( "Directory               = %s\n" % " ".join(dirs))
        readme.write(f"Jobfile                 = {job_file}\n\n")
        
        for i, infile in enumerate(cms_files):
            info = f"[{i+1:02d}] {infile}"
            print(info)
            readme.write(info+"\n")
        readme.write("\n\n")

        job.write(f'export CUDA_VISIBLE_DEVICES="{args.gpu_device}"\n\n')

        for n in range(args.start, args.start+args.repeat): 
            outdir = f"{args.prefix}{n:02d}"
            outdir_abspath = os.path.abspath(outdir)
            job.write(f"cd {outdir_abspath}/\n\n")
            if not os.path.exists(outdir):
                os.makedirs(outdir)
            msj_file = f"{outdir}/desmond_md_job_{n:02d}.msj"
            cfg_file = f"{outdir}/desmond_md_job_{n:02d}.cfg"
            cfg_file_basename= os.path.basename(cfg_file)
            
            with open(msj_file,"w") as msj:
                # modify desmond msj template
                md_msj.dot.simulate[-1].cfg_file = cfg_file_basename
                
                # Setting up restraints using the restraints keyword:
                # https://www.schrodinger.com/kb/332119
                if args.posres_force > 0.0:
                    # print the restraints in the multisim log file
                    md_msj.dot.simulate[-1].print_restraint = 'true'

                    # add the new terms defined in "restraints.new" to existing restraints.
                    # The default is restraints.existing = ignore which will 
                    # delete existing terms before adding any new ones.
                    # md_msj.dot.simulate[-1].restraints.existing = 'retain'

                    md_msj.dot.simulate[-1].restraints.new = [
                        {
                            'name'              : 'posre_harm',
                            'atoms'             : [ f'"{args.posres}"' ],
                            'force_constants'   : [ args.posres_force, ] * 3,
                        }
                        ]
                    # force constants in the x, y, and z direction

                md_msj.write(msj)

            with open(cfg_file,"w") as cfg:
                # modify desmond cfg template
                md_cfg.dot.randomize_velocity.seed = random.randint(1000, 9999)
                md_cfg.dot.time = total_simulation_time
                md_cfg.dot.temperature = t_schedule
                md_cfg.dot.trajectory.interval = args.interval
                md_cfg.write(cfg)

            for infile in cms_files:
                prefix_ = re.sub(r'desmond_setup[-_]','', os.path.basename(infile))
                prefix  = prefix_.replace("-out.cms","")
                
                job_name = f"desmond_md_job_{n:02d}_{prefix}"
                job.write('if [ ! -f {}/{} ]\n'.format( outdir_abspath, f"{job_name}-out.cms",))
                job.write('then\n')
                job.write('{} -JOBNAME {} -m {} -c {} -description "{}" {} {} -o {} -WAIT\n'.format(
                    multisim, 
                    job_name, 
                    os.path.basename(msj_file),
                    os.path.basename(cfg_file),
                    "GPU desmond MD",
                    opt,
                    os.path.join("..",infile),
                    f"{job_name}-out.cms",
                ))
                job.write('fi\n\n')

    os.chmod(job_file, 0o777)


import sys
import os
import argparse
import pandas as pd

try:
    from schrodinger.application.desmond.packages import analysis, topo, traj, traj_util
    from schrodinger.structutils import analyze
    from schrodinger import structure
except ImportError:
    print("Schrodinger Python API is required")
    sys.exit(0)



""" Analyze distance 

    The .cms files only contain physical atoms whereas the trajectory files may contain additional pseudo atoms. 
    This leads to two atom indices:
        Atom ID (AID) for structural files such as cms and mae files
            physical atoms only
            1-indexed
            results of ASL/SMARTS evaluation

    Global ID (GID) for trajectory files and Desmond MSYS files
            all atoms, both physical and pseudo
            0-indexed

Usage:

$SCHRODINGER/run ~/bucket/schrodinger/batch-desmond-distance.py \
    --receptor "C-555-SG" \
    --ligand "smarts. [CH3]"  \
    r0?/desmond_md_job_*-out.cms
"""

def get_atom_id(st:structure.Structure, chain: str, resSeq: int, name: str) -> int:
    """Get an aid with chain/resSeq/name.

    Args:
        st (structure.Structure): structure object
        chain (str): chainId
        resSeq (int): residue number
        name (str): atom name

    Returns:
        int: atom index
    """
    for a in st.atom:
        if (a.property['s_m_chain_name'] == chain) and \
            (a.property['i_m_residue_number'] == resSeq) and \
            (a.property['s_m_pdb_atom_name'].strip() == name) :
            return a.index
    return None



parser = argparse.ArgumentParser(
    description="trajectory distance",
    formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument('--atoms', dest="atoms", action="append", help="pair(s) of chain-resSeq-name separated by '-'")
parser.add_argument('--receptor', dest="receptor", default="", help="receptor chain-resSeq-name separated by '-'")
parser.add_argument('--ligand', dest="ligand", default="", help="ligand ASL")
parser.add_argument('--start', dest="start", type=float, default=0.0, help="start fraction for analysis")
parser.add_argument('--end', dest="end", type=float, default=1.0, help="end fraction for analysis")
parser.add_argument('--n', dest="n", type=int, default=0, help="target sample points")
parser.add_argument('--csv', dest="csv", default="distance.csv.gz", help="output csv or csv.gz filename")
parser.add_argument('--check', dest="check", default=False, action="store_true", help="just check atom index")
parser.add_argument('cms', nargs="+", help="desmond cms output file(s)")
args = parser.parse_args()


if __name__ == "__main__" :

    # check give atom representations
    distance_atom_pairs = []
    distance_names = []
    if not (args.receptor and args.ligand):
        if len(args.atoms) >= 2 and len(args.atoms) % 2 == 0:
            for i in range(0, len(args.atoms)//2):
                try:
                    (chain_1, resSeq_1, atomname_1) =  args.atoms[2*i+0].strip().split('-')
                    resSeq_1 = int(resSeq_1)
                except:
                    print("Error:", args.atoms[2*i+0])
                    print("each atom must be defined by {chain}-{reqSeq}-{atom_name} separated by '-'")
                    sys.exit(0)
                try:
                    (chain_2, resSeq_2, atomname_2) =  args.atoms[2*i+1].strip().split('-')
                    resSeq_2 = int(resSeq_2)
                except:
                    print("Error:", args.atoms[2*i+1])
                    print("each atom must be defined by {chain}-{reqSeq}-{atom_name} separated by '-'")
                    sys.exit(0)
                distance_name = args.atoms[2*i+0] + ':' + args.atoms[2*i+1]
                distance_names.append(distance_name)
                distance_atom_pairs.append((chain_1, resSeq_1, atomname_1, chain_2, resSeq_2, atomname_2, distance_name))

        else:
            print("Error: N x 2 atoms (or residues) are required to define distance(s)")
            sys.exit(0)

    else:
        # define distance_names
        distance_names = [ args.receptor + ':' + args.ligand ]

    data = {"name":[], "time":[], "atom_pair":[], "distance":[]}
    total_cms_files = len(args.cms)
    count_cms_files = 0
    
    print("distance name(s): ")
    for i, distance_name in enumerate(distance_names, start=1):
        print(f"    [{i}] {distance_name}")

    if not args.check:
        print("trajectory fractions : ", args.start, "-", args.end) 
    
    for input_cms in args.cms:
        # id
        name = os.path.basename(input_cms).replace("-out.cms","").replace("desmond_md_job_","")
        count_cms_files += 1

        # read .cms and its associated trajectory file
        # <jobname>-out.cms
        # <jobname>_trj

        if args.check:
            # only use cms
            (msys_model, cms_model) = topo.read_cms(input_cms)
        else:
            try:
                (msys_model, cms_model, trj) = traj_util.read_cms_and_traj(input_cms)
            except:
                (msys_model, cms_model) = topo.read_cms(input_cms)
                try:
                    trj_path = topo.find_traj_path(cms_model)  # or
                except:
                    try:
                        trj_path = topo.find_traj_path_from_cms_path(input_cms)
                    except:
                        trj_path = input_cms.replace("-out.cms","_trj")
                trj = traj.read_traj(trj_path)
        
            # number of frames
            num_frames = len(trj)
            startframe = max(int(num_frames * args.start), 0)
            endframe   = min(int(num_frames * args.end  ), num_frames)
            if args.n > 1:
                step = (endframe - startframe + 1) // args.n
            else:
                step = 1
            sys.stdout.write(f"[{count_cms_files}/{total_cms_files}] ({startframe}-{endframe})/{num_frames} {input_cms:60s}\r")
            sys.stdout.flush()

        
        list_of_obj = [] # distances to analyze
        list_of_tag = []

        if distance_atom_pairs:
            for (chain_1, resSeq_1, atomname_1, chain_2, resSeq_2, atomname_2, distance_name) in distance_atom_pairs:
                    aid_1 = get_atom_id(cms_model, chain_1, resSeq_1, atomname_1)
                    aid_2 = get_atom_id(cms_model, chain_2, resSeq_2, atomname_2)
                    # returns None unless four atom indexes are determined
                    if not (aid_1 and aid_2):
                        continue
                    if args.check:
                        print(input_cms)
                        print(f"    distance= {distance_name} aids= {aid_1} {aid_2}")
                        continue
                    obj = analysis.Distance(msys_model, cms_model, aid_1, aid_2)
                    list_of_obj.append(obj)
                    list_of_tag.append(distance_name)
        else:
            (rec_chain, rec_resSeq, rec_atomname) =  args.receptor.strip().split('-')
            rec_resSeq = int(rec_resSeq)
            aid_1 = get_atom_id(cms_model, rec_chain, rec_resSeq, rec_atomname)
            ligands = analyze.find_ligands(cms_model)
            aid_2 = analyze.evaluate_asl(cms_model, args.ligand)
            # specify substructure of the ligand
            aid_2 = list(set(aid_2) & set(ligands[0].atom_indexes))[0]
            # define list_of_obj, list_of_tag
            if args.check:
                print(input_cms)
                print(f"    distance= {distance_name} aids= {aid_1} {aid_2}")
                continue
            obj = analysis.Distance(msys_model, cms_model, aid_1, aid_2)
            list_of_obj.append(obj)
            list_of_tag.append(distance_name)
                
        if not args.check:
            
            list_of_results = analysis.analyze(trj[startframe:endframe:step], *list_of_obj)
            #list_of_results = analysis.analyze(trj, *list_of_obj)

            # analysis.analyze would return a list when one distance is calculated
            # analysis.analyze would return a list of list for more than one distances
            # for uniform process, let's make a list of list for one distance

            if len(list_of_tag) == 1:
                list_of_results = [ list_of_results ]

            for distance_name, results in zip(list_of_tag, list_of_results):
                num_data = len(results)
                data["name"] += num_data * [name,]
                data["time"] += [ fr.time for fr in trj[startframe:endframe:step] ] # (ps)
                data["atom_pair"] += num_data * [distance_name,]
                data["distance"] += results


    if not args.check:
        df = pd.DataFrame(data)
        df.to_csv(args.csv, sep=",", index=False, header=True, float_format="%.1f")
        print()
        print(f"dataframe written to {args.csv}")
        print()



import sys
import os
import argparse


try:
    from schrodinger import structure
    from schrodinger.structutils import analyze
    from schrodinger.application.desmond.packages import traj_util
except ImportError:
    print("Schrodinger Python API is required")
    sys.exit(0)

parser = argparse.ArgumentParser(description="trajectory information",
    formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument('--asl', dest="asl", default="", help="ASL")
parser.add_argument('cms', nargs="+", help="desmond cms output file(s)")
args = parser.parse_args()

for cms_file in args.cms:
    # id
    name = os.path.basename(cms_file).replace("-out.cms","").replace("desmond_md_job_","")

    # read .cms and its associated trajectory file
    (msys_model, cms_model, traj) = traj_util.read_cms_and_traj(cms_file)
    
    # number of frames
    num_frames = len(traj)

    # formal charge
    st = structure.StructureReader.read(cms_file)
    num_atoms = 0
    total_charge = 0
    for chain in st.chain:
        for res in chain.residue:
                for a in res.atom:
                    total_charge += int(a.property["i_m_formal_charge"])
                    num_atoms += 1

    print(f"{cms_file} frames {num_frames:4d} atoms {num_atoms} charge {total_charge}")

    if args.asl:
        atom_indice = analyze.evaluate_asl(st, args.asl)
        for i in atom_indice:
            print(i,
                st.atom[i].property["s_m_chain_name"],
                st.atom[i].property['i_m_residue_number'],
                st.atom[i].property["s_m_pdb_residue_name"],
                st.atom[i].property['s_m_pdb_atom_name'],
                st.atom[i].property["r_m_x_coord"],
                st.atom[i].property["r_m_y_coord"],
                st.atom[i].property["r_m_z_coord"],
                )


import os
import sys
import argparse
from desmondtools import Multisim


def batch_setup():
    SCHRODINGER = os.environ["SCHRODINGER"]
    multisim = "{}/utilities/multisim".format(SCHRODINGER)
    setup_msj = Multisim(template="desmond-setup.msj")
    
    parser = argparse.ArgumentParser(description="batch gdesmond md system setup",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-c','--conc', dest="conc", type=float, default=0.15, help="salt concentration in M")
    parser.add_argument('-d','--dist', dest="dist", type=float, default=10.0, help="buffer distance in A")
    parser.add_argument('-l','--lipid',  dest="lipid", default="", help="lipid bilayer")
    parser.add_argument('--cpp', dest="cpp", default=False, action="store_true", help="CPP simulation setup")
    parser.add_argument('-s','--solvent',  dest="solvent", default="TIP3P", help="solvent model")
    parser.add_argument('-i','--counterion',  dest="counterion", default="Na", help="neutralizing ion")
    parser.add_argument('-n','--negative', dest="neg", default="Cl", help="negative salt ion")
    parser.add_argument('-p','--positive', dest="pos", default="Na", help="positive salt ion")
    parser.add_argument('-f','--forcefield', dest="forcefield", default="S-OPLS", help="forcefield")
    parser.add_argument('-j','--jobfile', dest="job_file", default="desmond_setup_1.sh", help="job filename")
    parser.add_argument('-m','--msjfile', dest="msj_file", default="desmond_setup_1.msj", help="msj filename")
    parser.add_argument('-a','--appendix', dest="appendix", default="", help="job name appendix")
    parser.add_argument('mae', nargs="+", help="desmond mae file")
    args = parser.parse_args()

    if args.appendix:
        msj_file = args.msj_file[:-4] + "_" + args.appendix + ".msj"
    else:
        msj_file = args.msj_file

    if args.appendix:
        job_file = args.job_file[:-3] + "_" + args.appendix + ".sh"
    else:
        job_file = args.job_file

    # job file (.sh) and msj file (.msj) should match
    while os.path.exists(job_file):
        splited = job_file.replace(".sh","").split("_")
        splited[-1] = str(int(splited[-1]) + 1)
        job_file = "_".join(splited) + ".sh"

    while os.path.exists(msj_file):
        splited = msj_file.replace(".msj","").split("_")
        splited[-1] = str(int(splited[-1]) + 1)
        msj_file = "_".join(splited) + ".msj"

    with open(msj_file, "w") as msj:
        setup_msj.dot.build_geometry.solvent = str(args.solvent)
        setup_msj.dot.build_geometry.add_counterion.ion = str(args.counterion)
        setup_msj.dot.build_geometry.salt.concentration = str(args.conc)
        setup_msj.dot.build_geometry.box.size = f"[ {args.dist} {args.dist} {args.dist} ]"
        setup_msj.dot.build_geometry.salt.negative_ion = str(args.neg)
        setup_msj.dot.build_geometry.salt.positive_ion = str(args.pos)
        if args.cpp:
            setup_msj.dot.build_geometry.rezero_system = "False"
            args.lipid = "POPC"
        if args.lipid:
            setup_msj.dot.build_geometry.membrane_box.lipid = str(args.lipid)
        else:
            # remove membrane_box block
            setup_msj.dot.build_geometry.pop("membrane_box")
        setup_msj.dot.build_geometry.override_forcefield = str(args.forcefield)
        setup_msj.dot.assign_forcefield.forcefield = str(args.forcefield)
        setup_msj.dot.assign_forcefield.water = str(args.solvent)
        setup_msj.write(msj)


    with open("README","a") as readme, open(job_file,"w") as job:
        cmd_echo = ""
        for argv in sys.argv:
            if cmd_echo:
                cmd_echo += " "
            if " " in argv:
                cmd_echo += f'"{argv}"'
            else:
                cmd_echo += f'{argv}'
        readme.write(f"{cmd_echo}\n\n")
        readme.write(f"Force Field   = {args.forcefield}\n")
        readme.write(f"Solvent       = {args.solvent}\n")
        readme.write(f"Counter Ion   = {args.counterion}\n")
        readme.write(f"Positive Ion  = {args.pos}\n")
        readme.write(f"Negative Ion  = {args.neg}\n")
        readme.write(f"Concentration = {args.conc} (M)\n")
        readme.write(f"Size          = {args.dist} (A)\n")
        readme.write(f"Lipid         = {args.lipid}\n")
        readme.write(f"msjfile       = {msj_file}\n")
        readme.write(f"Jobfile       = {job_file}\n")
        readme.write(f"Input structure(s):\n")

        for i, infile in enumerate(args.mae):
            prefix = os.path.basename(infile).split(".")[0]
            job_name = f"desmond_setup-{prefix}"
            if args.appendix:
                job_name += f"-{args.appendix}"
            cms_file = f"{job_name}-out.cms"
            job.write(f"if [ ! -f {cms_file} ]\n")
            job.write(f"then\n")
            job.write(f"{multisim} \\\n")
            job.write(f"  -JOBNAME {job_name} \\\n")
            job.write(f"  -m {msj_file} {os.path.abspath(infile)} \\\n") 
            job.write(f"  -o {cms_file} \\\n")
            job.write(f"  -HOST localhost:20 -maxjob 20 -WAIT\n")
            job.write(f"fi\n")
            readme.write(f"[{i+1:02d}] {infile}\n")
        readme.write("\n\n")

    os.chmod(job_file, 0o777)


""" Analyze ligand pose stability """

try:
    from schrodinger.application.desmond.packages import analysis, topo, traj_util
except ImportError:
    print("schrodinger python API is required")
    sys.exit(0)

import sys
import os
import argparse
import pandas as pd
import numpy as np

# ASL ex. not (chain. D) and (not (atom.ele H)) 

parser = argparse.ArgumentParser(description="trajectory dihedral angles",
    formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument('--receptor', dest="receptor", default="mol. 1", 
    help="receptor ASL")
parser.add_argument('--ligand', dest="ligand", default="mol. 2", 
    help="ligand ASL")
parser.add_argument('--threshold', dest="threshold", type=float, default=2.0, 
    help="rmsd threshold")
parser.add_argument('--out', dest="out", default="ligand-rmsd.csv.gz", 
    help="output csv or csv.gz filename")
parser.add_argument('cms', nargs="+", help="desmond cms output file(s)")
args = parser.parse_args()

if __name__ == "__main__" :

    data = {
        "name" : [], 
        f"rmsd_percent({args.threshold:.1f}A)" : [], 
        "rmsd_median" : [], 
        "rmsd_mean" : [], 
        "rmsd_max" : [],
        "dist_init" : [], 
        "dist_median" : [], 
        "dist_mean" : [], 
        "dist_max" : [],
    }
    total_cms_files = len(args.cms)
    count_cms_files = 0

    # by default, heavy atom and no water
    receptor_asl = f"({args.receptor}) and (not atom.ele H) and (not water)"
    ligand_asl = f"({args.ligand}) and (not atom.ele H) and (not water)"

    logfile = args.out
    if logfile.endswith(".gz"):
        logfile = logfile[:-3]
    if logfile.endswith(".csv"):
        logfile = logfile[:-4]
    logfile = logfile + ".log"
    logging = open(logfile,"w")

    for input_cms in args.cms:

        (msys_model, cms_model, traj) = traj_util.read_cms_and_traj(input_cms)
        num_frames = len(traj)
        count_cms_files += 1

        name, ext_ = os.path.splitext(os.path.basename(input_cms))
        name = name.replace("desmond_md_job_","").replace("-out","")

        print(f"[{count_cms_files}/{total_cms_files}] {input_cms} ({num_frames} frames)")
        logging.write(f"[{count_cms_files}/{total_cms_files}] {input_cms} ({num_frames} frames)\n")

        receptor_aids = cms_model.select_atom(receptor_asl)
        print(f"\treceptor {len(receptor_aids)} atoms : {receptor_asl}")
        logging.write(f"\treceptor {len(receptor_aids)} atoms : {receptor_asl}\n")

        receptor_gids = topo.aids2gids(cms_model, receptor_aids, include_pseudoatoms=False)
        receptor_ref_pos = traj[0].pos(receptor_gids)

        ligand_aids = cms_model.select_atom(ligand_asl)
        print(f"\tligand {len(ligand_aids)} atoms : {ligand_asl}")
        logging.write(f"\tligand {len(ligand_aids)} atoms : {ligand_asl}\n")

        ligand_gids = topo.aids2gids(cms_model, ligand_aids, include_pseudoatoms=False)
        ligand_ref_pos = traj[0].pos(ligand_gids)
        
        # Ligand Root Mean Square Deviation from reference positions, 
        # with optional alignment fitting. 
        # Taking conformational symmetry into account.
        ligand_rmsd_analyzer = analysis.LigandRMSD(msys_model, cms_model, 
            aids = ligand_aids, 
            ref_pos = ligand_ref_pos,
            fit_aids = receptor_aids,
            fit_ref_pos = receptor_ref_pos,
            )

        receptor_centroid = analysis.Centroid(msys_model, cms_model, asl=receptor_asl)
        ligand_centroid = analysis.Centroid(msys_model, cms_model, asl=ligand_asl)
        ligand_dist_analyzer = analysis.Distance(msys_model, cms_model, 
            xid0 = receptor_centroid, 
            xid1 = ligand_centroid,
            )

        results = analysis.analyze(traj, ligand_rmsd_analyzer, ligand_dist_analyzer)
        # results[0] : list of rmsd
        # results[1] : list of centroid-to-centroid distance 
        
        rmsd_within_threshold = np.sum([1 for v in results[0] if v < args.threshold])
        rmsd_percent = 100.0 * rmsd_within_threshold / num_frames
        rmsd_mean_ = np.mean(results[0])
        rmsd_median_ = np.median(results[0])
        rmsd_max_ = np.max(results[0])
        dist_init_ = results[1][0]
        dist_mean_ = np.mean(results[1])
        dist_median_ = np.median(results[1])
        dist_max_ = np.max(results[1])
        print(f"\tLigandRMSD {rmsd_within_threshold} out of {num_frames} within {args.threshold:.1f}A ({rmsd_percent:.1f} %) ",end="")
        print(f"median {rmsd_median_:.2f} mean {rmsd_mean_:.2f} max {rmsd_max_:.2f}")
        print(f"\tCentroid Distance init {dist_init_:.2f} median {dist_median_:.2f} mean {dist_mean_:.2f} max {dist_max_:.2f}")
        print()
        logging.write(f"\tLigandRMSD {rmsd_within_threshold} out of {num_frames} within {args.threshold:.1f}A ({rmsd_percent:.1f} %) ")
        logging.write(f"median {rmsd_median_:.2f} mean {rmsd_mean_:.2f} max {rmsd_max_:.2f}\n")
        logging.write(f"\tCentroid Distance init {dist_init_:.2f} median {dist_median_:.2f} mean {dist_mean_:.2f} max {dist_max_:.2f}\n\n")
        
        data["name"].append(name)
        data[f"rmsd_percent({args.threshold:.1f}A)"].append(rmsd_percent)
        data["rmsd_median"].append(rmsd_median_)
        data["rmsd_mean"].append(rmsd_mean_)
        data["rmsd_max"].append(rmsd_max_)
        data["dist_init"].append(dist_init_)
        data["dist_median"].append(dist_median_)
        data["dist_mean"].append(dist_mean_)
        data["dist_max"].append(dist_max_)
    
    df = pd.DataFrame(data)
    df.to_csv(args.out, sep=",", index=False, header=True, float_format="%.3f")
    print(f"dataframe written to {args.out}")
    print()
    logging.write(f"dataframe written to {args.out}\n\n")



from .batch_desmond_mdinfo import batch_mdinfo
from .batch_desmond_cms2dcd import batch_cms2dcd
from .batch_desmond_setup import batch_setup
from .batch_desmond_md import batch_md
from .batch_desmond_min import batch_min
from .batch_desmond_metad import batch_metad
from .batch_desmond_mmgbsa import batch_mmgbsa
from .batch_desmond_pli import batch_pli

from .schrodinger_run import *


import sys
import os
import argparse
import pandas as pd

""" Analyze dihedral and pseudo-dihedrals """


try:
    from schrodinger.application.desmond.packages import analysis, topo, traj, traj_util
except ImportError:
    print("schrodinger python API is required")
    sys.exit(0)

# definition of dihedral angles

dihedral_def = {
    # standard protein backbone dihedral angles: phi, psi
    "phi"   : [(-1, "C"), (0, "N" ), (0, "CA"), (0, "C")],
    "psi"   : [( 0, "N"), (0, "CA"), (0, "C" ), (1, "N")],
    # standard nucleic acids backbone dihedral angles: alpha, beta, gamma, delta, epsilon, zeta
    "a"     : [(-1, "O3'"), (0, "P"  ), (0, "O5'"), (0, "C5'")],
    "b"     : [( 0, "P"  ), (0, "O5'"), (0, "C5'"), (0, "C4'")],
    "g"     : [( 0, "O5'"), (0, "C5'"), (0, "C4'"), (0, "C3'")],
    "d"     : [( 0, "C5'"), (0, "C4'"), (0, "C3'"), (0, "O3'")],
    "e"     : [( 0, "C4'"), (0, "C3'"), (0, "O3'"), (1, "P"  )],
    "z"     : [( 0, "C3'"), (0, "O3'"), (1, "P"  ), (1, "O5'")],
    # pseudo-dihedral angles
    "eta"   : [(-1, "C4'"), (0, "P"  ), (0, "C4'"), (1, "P"  )],
    "theta" : [( 0, "P"  ), (0, "C4'"), (1, "P"  ), (1, "C4'")],
    }



def get_dihedral_atom_ids(st, chain, resSeq, angle_name):
    if not angle_name in dihedral_def :
        return None

    residue_numbers = set()
    for (offset, atom_name) in dihedral_def[angle_name]:
        residue_numbers.add(offset+resSeq)
    atoms_ = []
    for a in st.atom:
        if (a.property['s_m_chain_name'] == chain) and \
            (int(a.property['i_m_residue_number']) in residue_numbers) :
            atoms_.append(a)

    atom_ids=[]
    for (offset, atom_name) in dihedral_def[angle_name]:
        for a in atoms_:
            if (int(a.property['i_m_residue_number']) == (resSeq+offset)) and \
                (a.property['s_m_pdb_atom_name'].strip() == atom_name) :
                atom_ids.append(a.index)
    
    if len(atom_ids) == 4 :
        return atom_ids
    
    return None


parser = argparse.ArgumentParser(description="trajectory dihedral angles",
    formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument('--chain', dest="chain", default="A", help="chain to be calculated")
parser.add_argument('--pseudo', dest="pseudo", default=False, action="store_true" , help="pseudo dihedral angles")
parser.add_argument('--eta', dest="eta", default=False, action="store_true" , help="pseudo dihedral angle eta")
parser.add_argument('--theta', dest="theta", default=False, action="store_true" , help="pseudo dihedral angle theta")
parser.add_argument('--nucleic', dest="nucleic", default=False, action="store_true" , help="nucleic acid dihedral angles")
parser.add_argument('--alpha', dest="alpha", default=False, action="store_true" , help="nucleic acid dihedral alpha")
parser.add_argument('--beta', dest="beta", default=False, action="store_true" , help="nucleic acid dihedral beta")
parser.add_argument('--gamma', dest="gamma", default=False, action="store_true" , help="nucleic acid dihedral gamma")
parser.add_argument('--delta', dest="delta", default=False, action="store_true" , help="nucleic acid dihedral delta")
parser.add_argument('--epsilon', dest="epsilon", default=False, action="store_true" , help="nucleic acid dihedral epsilon")
parser.add_argument('--zeta', dest="zeta", default=False, action="store_true" , help="nucleic acid dihedral zeta")
parser.add_argument('--protein', dest="protein", default=False, action="store_true" , help="protein backbone dihedral angles")
parser.add_argument('--phi', dest="phi", default=False, action="store_true" , help="protein backbone dihedral phi")
parser.add_argument('--psi', dest="psi", default=False, action="store_true" , help="protein backbone dihedral psi")
parser.add_argument('--resSeq_start', dest="resSeq_start", type=int, default=0, help="first resSeq for analysis")
parser.add_argument('--resSeq_end', dest="resSeq_end", type=int, default=0, help="last resSeq for analysis")
parser.add_argument('--traj_start', dest="traj_start", type=float, default=0.0, help="traj_start fraction for analysis")
parser.add_argument('--traj_end', dest="traj_end", type=float, default=1.0, help="end fraction for analysis")
parser.add_argument('--csv', dest="csv", default="dihedrals.csv.gz", help="output csv or csv.gz filename")
parser.add_argument('--check', dest="check", default=False, action="store_true", help="just check atom index")
parser.add_argument('cms', nargs="+", help="desmond cms output file(s)")
args = parser.parse_args()


if __name__ == "__main__" :
    # dihedral angles to be calculated
    names = []
    if args.nucleic :
        names.extend(["a", "b", "g", "d", "e", "z"])
    if args.alpha:
        names.append("a")
    if args.beta:
        names.append("b")
    if args.gamma:
        names.append("g")
    if args.delta:
        names.append("d")
    if args.epsilon:
        names.append("e")
    if args.zeta:
        names.append("z")
    if args.pseudo:
        names.extend(["eta", "theta"])
    if args.eta:
        names.extend("eta")
    if args.theta:
        names.append("theta")
    if args.protein:
        names.extend(["phi", "psi"])
    if args.phi:
        names.append("phi")
    if args.psi:
        names.append("psi")

    data = {"name":[], "resid":[], "angle_name":[], "angle":[]}
    total_cms_files = len(args.cms)
    count_cms_files = 0
    
    print("dihedral angles      : ", ",".join(names))
    if not args.check:
        print("trajectory fractions : ", args.traj_start, "-", args.traj_end) 
    
    for input_cms in args.cms:
        # id
        name = os.path.basename(input_cms).replace("-out.cms","").replace("desmond_md_job_","")
        count_cms_files += 1

        # read .cms and its associated trajectory file
        # <jobname>-out.cms
        # <jobname>_trj

        if args.check:
            # only use cms
            (msys_model, cms_model) = topo.read_cms(input_cms)

        else:
            try:
                (msys_model, cms_model, trj) = traj_util.read_cms_and_traj(input_cms)
            except:
                (msys_model, cms_model) = topo.read_cms(input_cms)
                try:
                    trj_path = topo.find_traj_path(cms_model)  # or
                except:
                    try:
                        trj_path = topo.find_traj_path_from_cms_path(input_cms)
                    except:
                        trj_path = input_cms.replace("-out.cms","_trj")
                trj = traj.read_traj(trj_path)

        
            # number of frames
            num_frames = len(trj)
            startframe = max(int(num_frames * args.traj_start), 0)
            endframe   = min(int(num_frames * args.traj_end  ), num_frames)
        
            sys.stdout.write(f"[{count_cms_files}/{total_cms_files}] ({startframe}-{endframe})/{num_frames} {input_cms:60s}\r")
            sys.stdout.flush()

        # residue numbers
        list_resSeq = []
        list_resName = []
        for c in cms_model.chain:
            if c.name == args.chain:
                for r in c.residue:
                    first_atom = list(r.atom)[0]
                    list_resSeq.append(int(first_atom.property['i_m_residue_number']))
                    list_resName.append(first_atom.property['s_m_pdb_residue_name'])
                # resNames = [ r.pdbres.strip() for r in c.residue ] # residue name
        
        list_resSeq = sorted(list_resSeq)
        resSeq_first = max(args.resSeq_start, list_resSeq[0])
        resSeq_last = min(args.resSeq_end, list_resSeq[-1])
        list_resSeq = [ x for x in list_resSeq if x >= resSeq_first and x <= resSeq_last ]

        list_of_obj = [] # dihedrals to analyze
        list_of_tag = []

        for angle_name in names:
            for resSeq in list_resSeq:
                dihed_aids = get_dihedral_atom_ids(cms_model, args.chain, resSeq, angle_name)
                # returns None unless four atom indexes are determined

                if not dihed_aids:
                    continue

                if args.check:
                    print(f"angle= {angle_name:8s} res= {resSeq:3d} aids= {dihed_aids}")
                    continue
                
                # convert atom index to global index
                #dihed_gids = topo.aids2gids(cms_model, dihed_aids, include_pseudoatoms=False)

                """
                print("angle_name=", angle_name, "resSeq=", resSeq)
                print("dihed_aids=", dihed_aids)
                print("    atom=", [cms_model.atom[x].property['s_m_pdb_atom_name'].strip() for x in dihed_aids])
                print("    res=", [cms_model.atom[x].property['i_m_residue_number'] for x in dihed_aids])
                print("dihed_gids=", dihed_gids)
                print("    atom=", [msys_model.atom(x).mass for x in dihed_gids])
                """
                
                # dihedral angle
                # * = unpacking argument list
                obj = analysis.Torsion(msys_model, cms_model, *dihed_aids)
                #print("frame_torsions=")
                #print([float(f"{v:.1f}") for v in analysis.analyze(trj[startframe:endframe], obj)])

                list_of_obj.append(obj)
                list_of_tag.append((angle_name, resSeq))
                #print()
                
        if not args.check:

            list_of_results = analysis.analyze(trj[startframe:endframe], *list_of_obj) 
            #print("list_of_results=", list_of_results[:20])
            # analysis.analyze returns a list of list

            for (angle_name, resSeq), results in zip(list_of_tag, list_of_results):
                #angle_in_degree = [v+360. if v < 0 else v for v in results] # 0...360
                angle_in_degree = [(v+360.0) % 360 for v in results] # 0...360
                num_data = len(angle_in_degree)
                data["name"]        += num_data * [name,]
                data["resid"]       += num_data * [resSeq,]
                data["angle_name"]  += num_data * [angle_name,]
                data["angle"]       += angle_in_degree


    if not args.check:
        df = pd.DataFrame(data)
        df.to_csv(args.csv, sep=",", index=False, header=True, float_format="%.1f")
        print()
        print(f"dataframe written to {args.csv}")
        print()



import sys
import io
import os
import errno
import importlib.resources
import pathlib
import pyparsing as pp
from pyparsing import pyparsing_common as ppc
from dotmap import DotMap


class Multisim:
    """Parsing Desmond multisim .cfg and .msj expressions"""
    # variable, value, array, expr
    # pyparsing module’s default behavior is to ignore whitespace.
    # +: AND, |: MatchFirst, left-to-right, ^: Or(longest match)
    # Group: --> list
    # Dict: --> dict
    # Forward: --> recursive

    EQ = pp.Suppress('=')
    LBRACKET, RBRACKET, LBRACE, RBRACE = map(pp.Literal, "[]{}")
    variable = (pp.Word(pp.alphanums + "._/?-@") + 
                pp.Opt("." + pp.Word(pp.alphanums))).set_parse_action(''.join)
    _string1 = pp.Word(pp.alphanums + "._/?-@*")
    _string2 = pp.quoted_string()
    _number  = ppc.number()
    value   = (_string1 | _string2 | _number)
    array   = pp.Forward()
    array   <<= pp.Group(LBRACKET + (pp.ZeroOrMore(value | array)) + RBRACKET)
    expr    = pp.Forward()
    _expr_0  = (variable + EQ + value)
    _expr_1  = (variable + EQ + array)
    _expr_2  = (variable + EQ + pp.Group(
        LBRACE + pp.ZeroOrMore(expr) + RBRACE ))
    _expr_3  = (variable + EQ + pp.Group(
        LBRACKET + pp.ZeroOrMore(pp.Group(LBRACE + expr + RBRACE)) + RBRACKET))
    _expr_4  = pp.Group(variable + pp.Group(
        LBRACE + pp.ZeroOrMore( expr ) + RBRACE))
    expr    <<= pp.OneOrMore(pp.Dict(pp.Group(
        _expr_0 | _expr_1 | _expr_2 | _expr_3 | _expr_4)))
    expr.ignore("#" + pp.restOfLine)


    def __init__(self, **kwargs):
        self.template_path = None
        self.ParseResults = None
        self.dict = {}
        self.dot = DotMap()
        self.output = sys.stdout # do not attempt to close
        self.indent = 4

        if 'template' in kwargs:
            template = kwargs['template']
            template_path = pathlib.Path(template)
            if template_path.is_file():
                self.template_path = template_path
            else:
                template_path = importlib.resources.files('desmondtools')
                self.template_path = template_path / template
            if self.template_path is None:
                raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), template)

            self.ParseResults = Multisim.expr.parse_file(self.template_path)
            self.decode()

        elif 'string' in kwargs:
            string = kwargs['string']
            try:
                self.ParseResults = Multisim.expr.parse_string(string)
                self.decode()
            except:
                raise RuntimeError("Multisim: cannot parse the input string.")
        else:
            raise RuntimeError("Multisim: template filename or string is required.")



    @staticmethod
    def unfold_dict_key(d) -> None:
        """Unfold '.' in the keys of a dictionary

        Args:
            d (dict): dictionary (to be modified)
        """
        ks = [ k for k in d ]
        for k in ks:
            v = d[k]
            if '.' in k:
                # d[k] = v --> d[kk[0]] = { kk[1] : v }
                _u = None
                _k = k.split('.')[0]
                for kk in k.split('.')[-1:0:-1]:
                    if _u is None:
                        _u = {kk : v }
                    else:
                        _u = {kk : _u}
                d[_k] = _u
                del d[k]
            else:
                pass
            if isinstance(v, dict):
                Multisim.unfold_dict_key(v)


    @staticmethod
    def traverse_dict(d) -> None:
        """Recursively traverse a nested dictionary/list

        Args:
            d (dict): dictionary (to be modified)
        """
        if isinstance(d, dict):
            for k,v in d.items():
                if isinstance(v, dict):
                    Multisim.traverse_dict(v)            
                elif isinstance(v, list):
                    if v == ['{','}']:
                        d[k] = {}
                    elif v == ['[',']']:
                        d[k] = []
                    elif v[0] == '[' and v[-1] == ']':
                        d[k] = v[1:-1]
                        Multisim.traverse_dict(d[k])
                    elif v[0] == '{' and v[-1] == '}':
                        d[k] = dict(v[1:-1])
                        Multisim.traverse_dict(d[k])
        elif isinstance(d, list):
            for k,v in enumerate(d):
                if isinstance(v, dict):
                    Multisim.traverse_dict(v)
                elif isinstance(v, list):
                    if v == ['{','}']:
                        d[k] = {}
                    elif v == ['[',']']:
                        d[k] = []
                    elif v[0] == '[' and v[-1] == ']':
                        d[k] = v[1:-1]
                        Multisim.traverse_dict(d[k])
                    elif v[0] == '{' and v[-1] == '}':
                        d[k] = dict(v[1:-1])
                        Multisim.traverse_dict(d[k])


    def decode(self) -> None:
        """decode the parsed results into a dictionary and its dotmap"""
        # create .dict
        if isinstance(self.ParseResults.as_list()[0][0], str): # key
            self.dict = self.ParseResults.as_dict()
            Multisim.traverse_dict(self.dict)
            # handle the case where key has '.'
            Multisim.unfold_dict_key(self.dict)

        elif isinstance(self.ParseResults.as_list()[0][0], list):
            self.dict = [] # now self.dict is a list of dictionary
            for section in self.ParseResults:
                dict_ = dict(section.as_list())
                Multisim.traverse_dict(dict_)
                # handle the case where key has '.'
                Multisim.unfold_dict_key(dict_)
                self.dict.append(dict_)

        # create .dot
        if isinstance(self.dict, list):
            self.dot = {}
            for d in self.dict:
                for k,v in d.items():
                    if k in self.dot:
                        if isinstance(self.dot[k], list):
                            self.dot[k].append(v)
                        else:
                            self.dot[k] = [self.dot[k], v]
                    else:
                        self.dot[k] = v
            self.dot = DotMap(self.dot)
        else:
            self.dot = DotMap(self.dict)

   

    def write(self, output=None):
        """Writes DOT object"""
        if isinstance(output, io.IOBase):
            self.output = output
        elif isinstance(output, str):
            self.output = open(output, "w")

        if isinstance(self.dict, list):
            blocks = []
            for k, v in self.dot.items():
                if isinstance(v, list):
                    for vv in v:
                        blocks.append({k:vv})
                else:
                    blocks.append({k:v})
            for block in blocks:
                self._write_dict(block, block=True)
                self.output.write("\n")
        else:
            self._write_dict(self.dot)


    def _write_dict(self, d, block=False, depth=0):
        """subroutine of .write() method"""
        spc = ' ' * self.indent * depth
        if isinstance(d, dict) or isinstance(d, DotMap):
            for k, v in d.items():
                k = str(k)
                if v:
                    if isinstance(v, dict) or isinstance(v, DotMap):
                        if depth == 0 and block:
                            self.output.write(spc + k + " {\n")
                        else:
                            self.output.write(spc + k + " = {\n")
                        self._write_dict(v, depth=depth+1)
                        self.output.write(spc + "}\n")
                    elif isinstance(v, list):
                        self.output.write(spc + k + " = [")
                        for vv in v:
                            if isinstance(vv, dict) or isinstance(vv, DotMap): 
                                self.output.write("{\n")
                            elif isinstance(vv, list):
                                self.output.write("[")
                            self._write_dict(vv, depth=depth+1)
                            if isinstance(vv, dict) or isinstance(vv, DotMap):
                                self.output.write(spc + "}")
                            elif isinstance(vv, list):
                                self.output.write("]")                
                        self.output.write("]\n")
                    else:
                        self.output.write(spc + k + " = " + str(v) +"\n")
                else:
                    if isinstance(v, list) and (not bool(v)):
                        self.output.write(spc + k + " = []\n")
                    elif (isinstance(v, dict) or isinstance(v, DotMap))and (not bool(v)):
                        self.output.write(spc + k + " = {\n}\n")
                    else:
                        self.output.write(spc + k + " =   \n")
        elif isinstance(d, list):
            for v in d:
                self._write_dict(v, depth=depth+1)
        else:
            self.output.write(" " + str(d) + " ")


    def to_dot(self) -> DotMap:
        """Returns parsed results as a DotMap object

        Returns:
            DotMap : DotMap object
        """
        return self.dot
    

    def to_list(self) -> list:
        """Returns parsed results as a list

        Returns:
            list : list
        """
        return self.ParseResults.as_list()


    def to_dict(self) -> dict:
        """Returns parsed results as a dictionary

        Returns:
            dict : dictionary
        """
        return self.dict



import os
import sys
import subprocess

from importlib.resources import files
script_path = files('desmondtools.script')

SCHRODINGER = os.environ["SCHRODINGER"]
USER=os.environ["USER"]

schrodinger_hosts = "{}/schrodinger.hosts".format(SCHRODINGER)
schrodinger_run = "{}/run".format(SCHRODINGER)
multisim = "{}/utilities/multisim".format(SCHRODINGER)

# Below python scripts were written with Schrodinger Python API
# and have be wrapped with $SCHRODINGER/run

def batch_cmsinfo() -> None:
    argv = sys.argv[1:]
    subprocess.run([schrodinger_run, 
                    script_path.joinpath('batch-desmond-cmsinfo.py')] + argv)

def batch_maeinfo() -> None:
    argv = sys.argv[1:]
    subprocess.run([schrodinger_run, 
                    script_path.joinpath('batch-desmond-maeinfo.py')] + argv)

def batch_report() -> None:
    argv = sys.argv[1:]
    subprocess.run([schrodinger_run, 
                    script_path.joinpath('batch-desmond-report.py')] + argv)

def batch_dihedral() -> None:
    argv = sys.argv[1:]
    subprocess.run([schrodinger_run, 
                    script_path.joinpath('batch-desmond-dihedral.py')] + argv)

def batch_distance() -> None:
    argv = sys.argv[1:]
    subprocess.run([schrodinger_run, 
                    script_path.joinpath('batch-desmond-distance.py')] + argv)

def batch_ligrmsd() -> None:
    argv = sys.argv[1:]
    subprocess.run([schrodinger_run, 
                    script_path.joinpath('batch-desmond-ligrmsd.py')] + argv)

def batch_rg() -> None:
    argv = sys.argv[1:]
    subprocess.run([schrodinger_run, 
                    script_path.joinpath('batch-desmond-rg.py')] + argv)

def batch_rmsx() -> None:
    argv = sys.argv[1:]
    subprocess.run([schrodinger_run, 
                    script_path.joinpath('batch-desmond-rmsx.py')] + argv)





import argparse
import subprocess
import os

from pathlib import Path


def batch_mmgbsa() -> None:
    SCHRODINGER = os.environ["SCHRODINGER"]
    schrodinger_run = "{}/run".format(SCHRODINGER)

    parser = argparse.ArgumentParser(description="trajectory thermal_mmgbsa",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--ligand', dest="ligand", default="chain. A", help="ASL for ligand")
    parser.add_argument('--solvent', dest="solvent", default="res. SPC T3P NA CL", help="ASL for solvent")
    parser.add_argument('--start', dest="start", type=int, help="start frame")
    parser.add_argument('--end', dest="end", type=int, help="end frame")
    parser.add_argument('--step', dest="step", type=int, default=1, help="frame step size")
    parser.add_argument('cms', nargs="+", help="desmond cms output file(s)")
    args = parser.parse_args()
    for cms_file in args.cms:
        jobname = Path(cms_file).name
        if jobname.startswith("desmond_md_job"):
            jobname = jobname.replace("desmond_md_job", "")
        if jobname.endswith("-out.cms"):
            jobname = jobname.replace("-out.cms", "")
        if not Path(f"{jobname}-prime-out.csv").exists():
            print(f"{cms_file} {args.start} {args.end} --> {jobname}")
            subprocess.run([schrodinger_run,
                            "thermal_mmgbsa.py",
                            "-j", jobname,
                            "-HOST", "localhost:10",
                            "-NJOBS", "2",
                            "-lig_asl", args.ligand,
                            "-atom_asl", args.solvent,
                            "-start_frame", str(args.start),
                            "-end_frame", str(args.end),
                            "-step_size", str(args.step),
                            cms_file,
                            ])
        else:
            print(f"{cms_file} {args.start} {args.end} --> Done/Skip")


import argparse
import subprocess
import desmondtools

from pathlib import Path
from importlib.resources import files


script_path = files('desmondtools.script')


def batch_cms2dcd() -> None:
    parser = argparse.ArgumentParser(description="Convert Desmond .cms to .dcd",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('cms', nargs="+", help="desmond cms output file(s)")
    args = parser.parse_args()
    for cms_file in args.cms:
        outdcd = Path(cms_file).stem + "-nowat.dcd"
        if not Path(outdcd).exists():
            subprocess.run(['vmd', '-display', 'text', 
                            '-e', script_path.joinpath('vmd_convert_cms_to_dcd.tcl'),
                            '-args', cms_file,
                            ])


import pytest

from desmondtools import Multisim


def test_Multisim_variable():
    # parse_string returns ['barostat.tau']
    o1 = Multisim.variable.parse_string("barostat.tau")[0]
    o2 = Multisim.variable.parse_string("barostat  .tau")[0]
    assert o1 == o2
    

def test_Multisim_expression_1():
    i = "task {} simulate { meta = [{a=1 b=3 c=[7 8 9]}] f = {} } simulate {n=2} simulate {n = 3}"
    D = Multisim(string=i).to_dot()
    assert D.simulate[0].meta[0].b == str(3)
    assert D.simulate[0].meta[0].c[1] == str(8)
    assert D.simulate[-1].n == str(3)


def test_Multisim_expression_2():
    i = "simulate { effect_if = [[ a ] b ] }"
    D = Multisim(string=i).to_dot()
    assert D.simulate.effect_if[0] == ['a']
    assert D.simulate.effect_if[1] == 'b'
    

def test_Multisim_expression_3():
    i = """ensemble = {
        brownie = {
            delta_max = 0.1
        }
        class = NVT
        method = Brownie
        thermostat = {
            tau = 1.0
        }
    }"""
    D = Multisim(string=i)
    assert D.dot.ensemble.brownie.delta_max == str(0.1)
    
    # assert D.dot.ensemble.class == 'NVT'
    #                       ^^^^^
    # SyntaxError: invalid syntax
    # 'class' is reserved word in python and raises SyntaxError
    # use instead .['class']
    assert D.dot.ensemble['class'] == 'NVT' # <--- it does not raise SyntaxError

    assert D.dot.ensemble.method == 'Brownie'
    assert D.dot.ensemble.thermostat.tau == str(1.0)
    D.write()


def test_Multisim_expression_4():
    i = """task {
        task = "desmond:auto"
    }
    simulate {
        title = "NPT and no restraints, 24ps"
        effect_if = [[ "@*.*.annealing" ] 'annealing = off temperature = "@*.*.temperature[0][0]"' ]
        time = 24
        ensemble = {
            class = NPT
            method = Langevin
            thermostat.tau = 0.1
            barostat.tau = 2.0
        }
        eneseq.interval = 0.3
        trajectory.center = solute
    }
    simulate {
        meta = {
            cv = [
                { 
                    atom = [0 0 0 0] 
                    type = dihedral 
                    wall = 0  
                    width = 5.0 
                }
            ]
            cv_name = "$JOBNAME$[_replica$REPLICA$].cvseq"
            first = 0.0
            height = 0.03
            interval = 0.09
            name = "$JOBNAME$[_replica$REPLICA$].kerseq"
        }
    }
    """

    D = Multisim(string=i)
    
    assert len(D.dot.simulate[0].effect_if) == 2
    assert len(D.dot.simulate[0].effect_if[0]) == 1
    assert D.dot.simulate[0].ensemble.method == 'Langevin'
    assert D.dot.simulate[0].ensemble.barostat.tau == str(2.0)
    assert D.dot.simulate[0].eneseq.interval == str(0.3)
    assert D.dot.simulate[0].trajectory.center == 'solute' # <---
    
    assert len(D.dot.simulate[1].meta.cv) == 1
    assert len(D.dot.simulate[1].meta.cv[0].atom) == 4
    assert D.dot.simulate[1].meta.height == str(0.03)
    assert D.dot.simulate[-1].meta.cv[0].wall == str(0)
    
    D.dot.simulate[0].ensemble.barostat.tau = 2.1
    assert D.dot.simulate[0].ensemble.barostat.tau == 2.1

    D.dot.simulate[1].meta.height = 0.06
    assert D.dot.simulate[1].meta.height == 0.06
    
    D.dot.simulate[-1].meta.cv[0].wall = 50.0
    assert D.dot.simulate[-1].meta.cv[0].wall == 50.0
    

def test_Multisim_expression_5():
    D= Multisim(template="desmond-md.msj")
    D.write()


import sys
import os
import argparse
import pandas as pd

try:
    from schrodinger import structure
except ImportError:
    print("Schrodinger Python API is required")
    sys.exit(0)

parser = argparse.ArgumentParser(description="Trajectory RMSD and RMSF",
    formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument('--eaf', dest="eaf", default="", help="event analysis outfile")
parser.add_argument('--prefix', dest="prefix", default="", help="output prefix")
parser.add_argument('cms', help="desmond cms output file")
args = parser.parse_args()

st_reader = structure.StructureReader(args.cms)
st = next(st_reader)

tbl = {}
rmsd_data = {'time_ns':[], 'rmsd':[]}
rmsf_data = {"resSeq":[], "resName":[], "name":[], "atom_index":[], "rmsf":[]}

with open(args.eaf, "r") as f:
    for line in f:
        line = line.strip()
        if "Name" in line:
            name = line.split()[2]
            tbl[name] = None
        
        if "Result" in line:
            data = line.split("=")[1]
            data = data.replace("[","")
            data = data.replace("]","")
            data = data.split()
            tbl[name] = data
        
        if "AtomIndices" in line:
            data = line.split("=")[1]
            data = data.replace("[","")
            data = data.replace("]","")
            data = data.split()
            tbl["AtomIndices"] = data
        
        if "TrajectoryFirstTime" in line:
            TrajectoryFirstTime = float(line.split()[2])
        
        if "TrajectoryInterval_ps" in line:
            TrajectoryInterval_ps = float(line.split()[2])
        
        if "TrajectoryNumFrames" in line:
            TrajectoryNumFrames = int(line.split()[2])
    
    for i, x in enumerate(tbl["RMSD_A"]):
        rmsd_str = f"{float(x):.3f}"
        rmsd_data['time_ns'].append(f"{TrajectoryFirstTime + i*TrajectoryInterval_ps*0.001:.3f}")
        rmsd_data['rmsd'].append(rmsd_str)
    
    for i, x in enumerate(tbl["RMSF_A"]):
        a = st.atom[int(tbl["AtomIndices"][i])]
        rmsf_str = f"{float(x):.3f}"
        rmsf_data['resSeq'].append(a.resnum)
        rmsf_data['resName'].append(a.pdbres.strip())
        rmsf_data['name'].append(a.pdbname.strip())
        rmsf_data['atom_index'].append(a.index)
        rmsf_data['rmsf'].append(rmsf_str)

rmsd_df = pd.DataFrame(rmsd_data)
rmsf_df = pd.DataFrame(rmsf_data).sort_values(by=['resSeq','resName','atom_index'])

rmsd_df.to_csv(f"{args.prefix}-rmsd.csv", index=False)
rmsf_df.to_csv(f"{args.prefix}-rmsf.csv", index=False)


from pathlib import Path
from dotmap import DotMap
from desmondtools import Multisim



class Event:
    def __init__(self, filename:Path | str):
        if isinstance(filename, Path):
            result = Multisim.expr.parse_file(filename.as_posix())
        else:
            result = Multisim.expr.parse_file(filename)
        d = result.as_dict()
        Multisim.traverse_dict(d)
        self.dot = DotMap(d)
        


class Interaction(Event):
    def __init__(self, filename:Path | str, verbose:bool=False):
        super().__init__(filename)

        self.num_frames = 0
        self.HBond = {}
        self.Hydrophobic = {}
        self.WaterBridge = {}
        self.Polar = {}
        self.HalogenBond = {}
        self.LigWat = {}
        self.Metal = {}
        self.PiCat = {}
        self.PiPi = {}

        for section in self.dot.Keywords:
            try:
                assert section.ProtLigInter.HBondResult
                self.num_frames = len(section.ProtLigInter.HBondResult)
                for frame in section.ProtLigInter.HBondResult:
                    # [[3 "_:ARG_143:HH22" d-s "L-FRAG_0:N6" ]]
                    for (frameno, prot, hbond_type, lig) in frame:
                        prot = prot.strip('\"')
                        (_, resid, atom) = prot.split(":")
                        (resName, resSeq) = resid.split("_")
                        resSeq = int(resSeq)
                        if resSeq in self.HBond:
                            self.HBond[resSeq]['count'] += 1
                        else:
                            self.HBond[resSeq] = {'resName': resName, 'count':1 }
                for resSeq in sorted(self.HBond):
                    fraction = float(self.HBond[resSeq]['count'])/self.num_frames
                    if verbose:
                        print(f"HBond {self.HBond[resSeq]['resName']}_{resSeq} {fraction:5.3f} {self.num_frames}")
            except:
                pass

            try:
                assert section.ProtLigInter.HydrophobicResult
                self.num_frames = len(section.ProtLigInter.HydrophobicResult)
                for frame in section.ProtLigInter.HydrophobicResult:
                    # [[0 "_:PHE_223" L-FRAG_0 ] [0 "_:ALA_241" L-FRAG_0 ]]
                    for (frameno, prot, lig) in frame:
                        prot = prot.strip('\"')
                        (_, resid) = prot.split(":")
                        (resName, resSeq) = resid.split("_")
                        resSeq = int(resSeq)
                        if resSeq in self.Hydrophobic:
                            self.Hydrophobic[resSeq]['count'] += 1
                        else:
                            self.Hydrophobic[resSeq] = {'resName': resName, 'count':1 }
                for resSeq in sorted(self.Hydrophobic):
                    fraction = float(self.Hydrophobic[resSeq]['count'])/self.num_frames
                    if verbose:
                        print(f"Hydrophobic {self.Hydrophobic[resSeq]['resName']}_{resSeq} {fraction:5.3f} {self.num_frames}")
            except:
                pass
            
            try:
                assert section.ProtLigInter.PolarResult
                self.num_frames = len(section.ProtLigInter.PolarResult)
                for frame in section.ProtLigInter.PolarResult:
                    # [[1 "_:GLU_216:OE2" b "L-FRAG_1:N3" 4.45 ]]
                    for (frameno, prot, _, lig, _) in frame:
                        prot = prot.strip('\"')
                        (_, resid, atom) = prot.split(":")
                        (resName, resSeq) = resid.split("_")
                        resSeq = int(resSeq)
                        if resSeq in self.Polar:
                            self.Polar[resSeq]['count'] += 1
                        else:
                            self.Polar[resSeq] = {'resName': resName, 'count':1 }
                for resSeq in sorted(self.Polar):
                    fraction = float(self.Polar[resSeq]['count'])/self.num_frames
                    if verbose:
                        print(f"Polar {self.Polar[resSeq]['resName']}_{resSeq} {fraction:5.3f} {self.num_frames}")
            except:
                pass

            try:
                assert section.ProtLigInter.WaterBridgeResult
                self.num_frames = len(section.ProtLigInter.WaterBridgeResult)
                for frame in section.ProtLigInter.WaterBridgeResult:
                    # [[3 "_:GLU_216:OE2" a "L-FRAG_0:N2" a 2431 ]]
                    for (frameno, prot, _, lig, _, _) in frame:
                        prot = prot.strip('\"')
                        (_, resid, atom) = prot.split(":")
                        (resName, resSeq) = resid.split("_")
                        resSeq = int(resSeq)
                        if resSeq in self.WaterBridge:
                            self.WaterBridge[resSeq]['count'] += 1
                        else:
                            self.WaterBridge[resSeq] = {'resName': resName, 'count':1 }
                for resSeq in sorted(self.WaterBridge):
                    fraction = float(self.WaterBridge[resSeq]['count'])/self.num_frames
                    if verbose:
                        print(f"WaterBridge {self.WaterBridge[resSeq]['resName']}_{resSeq} {fraction:5.3f} {self.num_frames}")
            except:
                pass


import sys
import os
import argparse

# entry > molecule > chain > residue > atom

try:
    from schrodinger import structure
    from schrodinger.structutils import analyze
    from schrodinger.application.desmond.packages import traj_util
except ImportError:
    print("Schrodinger Python API is required")
    sys.exit(0)

parser = argparse.ArgumentParser(description="mae information",
    formatter_class=argparse.ArgumentDefaultsHelpFormatter)
# parser.add_argument('--receptor', dest="receptor", default="", help="ASL for receptor")
# parser.add_argument('--ligand', dest="ligand", default="", help="ASL for ligand")
parser.add_argument('mae', nargs="+", help="mae or maegz file(s)")
args = parser.parse_args()

# examples:
#   receptor_asl = '(chain.name "A" AND res.num 833 AND atom.name "SG")'
#   ligand_asl = 'smarts. [N]C(=O)[OH1]'

indent = 4

for mae_file in args.mae:
    for st_index, st in enumerate(structure.StructureReader(mae_file), start=1):
        try:
            title = st.property['s_m_title']
        except:
            title = None
        try:
            entry_id = st.property['s_m_entry_id']
        except:
            entry_id = None
        try:
            entry_name = st.property['s_m_entry_name']
        except:
            entry_name = None

        print(f"(mol. {st_index}) {title}")
        for chain in st.chain:
            print(" " * indent, end='')
            print(f"(chain. {chain.name})")
            residue_numbers = None
            residue_names = {}
            for residue in chain.residue:
                first_atom = list(residue.atom)[0]
                resSeq = first_atom.property['i_m_residue_number']
                resName = first_atom.property['s_m_pdb_residue_name']
                if residue_numbers is None:
                    residue_numbers = [ resSeq ]
                else:
                    residue_numbers.append( resSeq )
                residue_names[resSeq] = resName

            start = None
            last = None
            continous_residues = []
            for resSeq in sorted(residue_numbers):
                if start is None:
                    start = resSeq
                if (last is not None) and (resSeq-last) != 1 :
                    continous_residues.append((start, last))
                    start = resSeq
                last = resSeq
            continous_residues.append((start, last))

            for i, (start, end) in enumerate(continous_residues):
                if i > 0:
                    (prev_start, prev_end) = continous_residues[i-1]
                    print(" " * indent*2, end='')
                    print(f"{prev_end+1:4d} .. {start-1:4d} ({start-prev_end-1}) * missing *")    
                print(" " * indent*2, end='')
                print(f"{residue_names[start]}{start:4d} .. ", end="")
                print(f"{residue_names[end]}{end:4d} ({end-start+1})")
        print()

    # receptor_atoms = []
    # ligands = analyze.find_ligands(st)
    # print(f"number of ligands: {len(ligands)}")
    # num_entry = 0
    # for mol in st:
    #     if not receptor_atoms :
    #         receptor_atoms = analyze.evaluate_asl(mol, args.receptor)
    #         continue
    #     # ligand entries
    #     ligand_atoms = analyze.evaluate_asl(mol, args.ligand)
    #     print(receptor_atoms, ligand_atoms)
    # print("total", num_entry,"molecules")


import os
import re
import sys
import glob
import argparse

from datetime import timedelta, datetime    


def batch_mdinfo():
    SCHRODINGER = os.environ["SCHRODINGER"]
    USER=os.environ["USER"]
    schrodinger_hosts = "{}/schrodinger.hosts".format(SCHRODINGER)

    simulation_time = re.compile(r'last_time = "(?P<t>[.0-9]+)"')
    #    last_time = "100000.0"
    progress = re.compile(r'Chemical time:\s+(?P<finished>[.0-9]+) ps, Step:\s+[0-9]+, ns/day:\s+(?P<rate>[.0-9]+)')
    #Chemical time:         33507.6000 ps, Step: 5584600, ns/day:      296.231

    parser = argparse.ArgumentParser(description="desmond MD info.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-t', '--tmpdir', dest='tmpdir', default=None, help='desmond temporary directory')
    parser.add_argument('-l', '--logfile', dest='logfile', default=None)
    args = parser.parse_args()

    if args.tmpdir is None:
        with open(schrodinger_hosts, "r") as f:
            for line in f:
                if line.strip().startswith("#"):
                    continue
                c = line.strip().split(":")
                if len(c) != 2: 
                    continue
                k = c[0].strip()
                v = c[1].strip()
                if k == 'tmpdir':
                    args.tmpdir = os.path.join(v, USER)
            if not os.path.exists(args.tmpdir):
                sys.stderr.write("Use -t or --tmpdir to give temporary directory\n")
                sys.exit(1)


    if args.logfile is None:
        args.logfile = args.tmpdir+"/desmond_*/*.log"
        logfile = None
        for f in glob.glob(args.logfile):
            if not "multisim" in f:
                logfile = f
                break
        if not logfile:
            sys.stderr.write(f"desmond log not found in {args.logfile}\n")
            sys.stderr.write("Use -l or --logfile to give log file path\n")
            sys.stderr.write("Use -t or --tmpdir to give temporary directory\n")
            sys.exit(2)


    total = None
    rate = []
    with open(logfile,"r") as f:
        for line in f:
            m = simulation_time.search(line)
            n = progress.search(line)
            if not total and m and m.group("t"):
                total = float(m.group("t"))*0.001
            if n and n.group("finished") and n.group("rate"):
                rate.append(float(n.group("rate")))
                finished = float(n.group("finished"))*0.001

        n = len(rate)
        print("-"*80)
        print(f"| SCHRODINGER= {SCHRODINGER}")
        print(f"| tmpdir: {args.tmpdir}")
        print(f"|")
        print(f"| Desmond MD Timing Info")
        print(f"| ----------------------")
        print(f"|")
        print(f"| {os.path.dirname(logfile)}/")
        print(f"| {os.path.basename(logfile)}")
        print(f"|")

        if n == 0:
            if total:
                print(f"| Total     {total:9.2f} ns")
            print(f"| Timing data not available yet")
            print("-"*80)
            sys.exit(0)

        remaining = total-finished
        avg_rate = sum(rate[-n:])/n
        eta = 24.0*remaining/avg_rate # hours
        eta_time = datetime.now() + timedelta(hours=eta)
        print(f"| Total     {total:9.2f} ns")
        print(f"| Completed {finished:9.2f} ns ({100.0*finished/total:5.1f}%)")
        print(f"| Remaining {remaining:9.2f} ns")
        print(f"| ")
        print(f"| Average timings")
        print(f"|    ns/day = {avg_rate:9.2f}")
        print(f"| ")
        print(f"| Estimated time")
        print(f"|    remaining = {eta:.2f} hours")
        print(f"|    completed = {eta_time.ctime()}")
        print("-"*80)


import argparse
import subprocess
import os

from desmondtools import Multisim

from importlib.resources import files

import desmondtools
script_path = files('desmondtools.script')


SCHRODINGER = os.environ["SCHRODINGER"]

USER=os.environ["USER"]

schrodinger_hosts = "{}/schrodinger.hosts".format(SCHRODINGER)
schrodinger_run = "{}/run".format(SCHRODINGER)

multisim = "{}/utilities/multisim".format(SCHRODINGER)

def batch_rmsx() -> None:
    parser = argparse.ArgumentParser(description="trajectory RMSD and RMSF",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--protein', dest="protein", default="(mol. 1) and not atom.ele H", help="ASL for protein")
    parser.add_argument('--ligand', dest="ligand", default="(mol. 2) and not atom.ele H", help="ASL for ligand")
    parser.add_argument('cms', nargs="+", help="desmond cms output file(s)")
    args = parser.parse_args()
    event_analysis_infile  = "temporary-event-in.eaf"
    event_analysis_outfile = "temporary-event-out.eaf"
    events = Multisim(template="desmond-rmsx-analysis.eaf")
    events.dot.Keywords[0].RMSD.ASL = args.protein
    events.dot.Keywords[1].RMSF.ASL = args.ligand
    events.dot.Keywords[1].RMSF.FitBy = args.protein
    events.dot.ProteinASL = args.protein
    events.dot.LigandASL = args.ligand
    for cms_file in args.cms:
        print(f"analyzing {cms_file} ...")
        prefix = os.path.basename(cms_file).replace('-out.cms','')
        trj = cms_file.replace("-out.cms", "_trj")
        events.dot.Trajectory = cms_file
        with open(event_analysis_infile, "w") as f:
            events.write(f)
        # requires Schrodinger Python API
        subprocess.run([schrodinger_run, 
                        "analyze_simulation.py",
                        cms_file, 
                        trj, 
                        event_analysis_outfile, 
                        event_analysis_infile],
                        stdout=subprocess.DEVNULL, 
                        stderr=subprocess.STDOUT)
        if os.path.exists(event_analysis_infile):
            os.remove(event_analysis_infile)
        # requires Schrodinger Python API
        subprocess.run([schrodinger_run, 
                        script_path.joinpath('batch-desmond-rmsx.py'),
                        cms_file, 
                        event_analysis_outfile, 
                        prefix])
        if os.path.exists(event_analysis_outfile):
            os.remove(event_analysis_outfile)


import sys
import os
import argparse
import pandas as pd

""" Analyze radius of gyration """


try:
    from schrodinger.application.desmond.packages import analysis, topo, traj, traj_util
    from schrodinger.structutils import analyze
except ImportError:
    print("schrodinger python API is required")
    sys.exit(0)


def check_asl_atoms(st, asl):
    # check ASL
    aids = analyze.evaluate_asl(st, asl)
    # collect chain, resSeq, resName, atomnames matching the ASL
    chain_ = set()
    resSeq_ = set()
    resName_ = set()
    # atomName_ = set()
    for a in st.atom:
        if a.index in aids:
            chain_.add(a.property['s_m_chain_name'])
            resSeq_.add(str(a.property['i_m_residue_number']))
            resName_.add(a.property['s_m_pdb_residue_name'].strip())
            # atomName_.add(a.property['s_m_pdb_atom_name'].strip())
    print("  asl: ", asl, "(number of atoms= ", len(aids),")")
    print("  chain(s): " + " ".join(chain_))
    print("  residue number(s): " + " ".join(resSeq_))
    print("  residue name(s): " + " ".join(resName_))
    # print("atom name(s): " + " ".join(atomName_))


parser = argparse.ArgumentParser(description="trajectory Rg",
    formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument('--asl', dest="asl", action="append", help="atom selection language(s)")
parser.add_argument('--start', dest="start", type=float, default=0.0, help="start fraction for analysis")
parser.add_argument('--end', dest="end", type=float, default=1.0, help="end fraction for analysis")
parser.add_argument('--n', dest="n", type=int, default=0, help="target sample points")
parser.add_argument('--csv', dest="csv", default="rg.csv.gz", help="output csv or csv.gz filename")
parser.add_argument('--check', dest="check", default=False, action="store_true", help="just check atom index")
parser.add_argument('cms', nargs="+", help="desmond cms output file(s)")
args = parser.parse_args()


if __name__ == "__main__" :
    data = {"name":[], "asl":[], "Rg":[]}
    total_cms_files = len(args.cms)
    count_cms_files = 0

    if not args.check:
        print("trajectory fractions : ", args.start, "-", args.end) 
    
    for input_cms in args.cms:
        # id
        name = os.path.basename(input_cms).replace("-out.cms","").replace("desmond_md_job_","")
        count_cms_files += 1

        # read .cms and its associated trajectory file
        # <jobname>-out.cms
        # <jobname>_trj

        if args.check:
            # only use cms
            (msys_model, cms_model) = topo.read_cms(input_cms)
        else:
            try:
                (msys_model, cms_model, trj) = traj_util.read_cms_and_traj(input_cms)
            except:
                (msys_model, cms_model) = topo.read_cms(input_cms)
                try:
                    trj_path = topo.find_traj_path(cms_model)  # or
                except:
                    try:
                        trj_path = topo.find_traj_path_from_cms_path(input_cms)
                    except:
                        trj_path = input_cms.replace("-out.cms","_trj")
                trj = traj.read_traj(trj_path)
        
            # number of frames
            num_frames = len(trj)
            startframe = max(int(num_frames * args.start), 0)
            endframe   = min(int(num_frames * args.end  ), num_frames)
            if args.n > 1:
                step = (endframe - startframe + 1) // args.n
            else:
                step = 1
            sys.stdout.write(f"[{count_cms_files}/{total_cms_files}] ({startframe}-{endframe})/{num_frames} {input_cms:60s}\r")
            sys.stdout.flush()

        list_of_obj = []
        list_of_tag = []
        for asl in args.asl:
            if args.check:
                print(input_cms)
                check_asl_atoms(cms_model, asl)
                continue
            obj = analysis.Gyradius(msys_model, cms_model, asl=asl)
            list_of_obj.append(obj)
            list_of_tag.append(asl)

        if not args.check:
            list_of_results = analysis.analyze(trj[startframe:endframe:step], *list_of_obj)
            # analysis.analyze would return a list when one distance is calculated
            # analysis.analyze would return a list of list for more than one distances
            # for uniform process, let's make a list of list for one distance
            if len(list_of_tag) == 1:
                list_of_results = [ list_of_results ]

            for tag, results in zip(list_of_tag, list_of_results):
                num_data = len(results)
                data["name"]    += num_data * [name,]
                data["asl"]     += num_data * [tag,]
                data["Rg"]      += results


    if not args.check:
        df = pd.DataFrame(data)
        df.to_csv(args.csv, sep=",", index=False, header=True, float_format="%.1f")
        print()
        print(f"dataframe written to {args.csv}")
        print()



import argparse
import sys
import pandas as pd

from desmondtools import Interaction


def batch_pli():
    teststring ="""Keywords = [
    {RMSD = {
        ASL = "((mol. 1 and backbone) and not (atom.ele H) and not (mol. 2))"
        Frame = 0
        Panel = pl_interact_survey
        Result = [0.0 1.161 1.286 1.331 1.176 1.195 ]
        SelectionType = Backbone
        Tab = pl_rmsd_tab
        Type = ASL
        Unit = Angstrom
        }
    }
    {RMSD = {
        ASL = "(mol. 1 and sidechain and not (mol. 2))"
        Frame = 0
        Panel = pl_interact_survey
        Result = [0.0 1.161 1.286 1.331 1.176 1.195 ]
        SelectionType = "Side chains"
        Tab = pl_rmsd_tab
        Type = ASL
        Unit = Angstrom
        }
    }
    ]"""
    # result = expr.parse_string(teststring)
    # d = result.as_dict()
    # traverse_dict(d)
    # dot = DotMap(d)
    # try:
    #     assert dot.Keywords[1].RMSD.SelectionType == '"Side chains"'
    #     print("ok")
    # except:
    #     print("error")

    parser = argparse.ArgumentParser(description="Average Protein-Ligand Interactions",
    formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--out', dest='out', default='mean-PLIF', help="output basename")
    parser.add_argument('eaf', nargs='+', default=[], help='input -out.eaf filename(s)')
    args = parser.parse_args()

    if len(args.eaf) == 0:
        argparse.print_help()
        sys.exit(0)

    total_num_frames = 0
    data = {}
    for filename in args.eaf:
        event_analysis = Interaction(filename, verbose=False)
        total_num_frames += event_analysis.num_frames
        
        print(f"{filename}  {event_analysis.num_frames} frames")
        
        for resSeq in event_analysis.HBond:
            if resSeq in data:
                data[resSeq]['hbond'] += event_analysis.HBond[resSeq]['count']
            else:
                data[resSeq] = {'resName': event_analysis.HBond[resSeq]['resName'], 
                                'hbond': event_analysis.HBond[resSeq]['count'],
                                'hydrophobic': 0,
                                'polar': 0,
                                'waterbridge': 0,
                                }
        for resSeq in event_analysis.Hydrophobic:
            if resSeq in data:
                data[resSeq]['hydrophobic'] += event_analysis.Hydrophobic[resSeq]['count']
            else:
                data[resSeq] = {'resName': event_analysis.Hydrophobic[resSeq]['resName'], 
                                'hbond': 0,
                                'hydrophobic': event_analysis.Hydrophobic[resSeq]['count'],
                                'polar': 0,
                                'waterbridge': 0,
                                }
        for resSeq in event_analysis.Polar:
            if resSeq in data:
                data[resSeq]['polar'] += event_analysis.Polar[resSeq]['count']
            else:
                data[resSeq] = {'resName': event_analysis.Polar[resSeq]['resName'], 
                                'hbond': 0,
                                'hydrophobic': 0,
                                'polar': event_analysis.Polar[resSeq]['count'],
                                'waterbridge': 0,
                                }
        for resSeq in event_analysis.WaterBridge:
            if resSeq in data:
                data[resSeq]['waterbridge'] += event_analysis.WaterBridge[resSeq]['count']
            else:
                data[resSeq] = {'resName': event_analysis.WaterBridge[resSeq]['resName'],
                                'hbond': 0,
                                'hydrophobic' : 0,
                                'polar': 0,
                                'waterbridge': event_analysis.WaterBridge[resSeq]['count'],
                                } 

    csvdata = {'resid':[], 
            'resSeq':[], 
            'resName':[], 
            'hbond':[], 
            'hydrophobic':[],
            'polar': [],
            'waterbridge': [],
            }
    
    for resSeq in sorted(data):
        csvdata['resSeq'].append(resSeq)
        csvdata['resName'].append(data[resSeq]['resName'])
        csvdata['resid'].append(f"{data[resSeq]['resName']}_{resSeq}")
        csvdata['hbond'].append(float(data[resSeq]['hbond'])/total_num_frames)
        csvdata['hydrophobic'].append(float(data[resSeq]['hydrophobic'])/total_num_frames)
        csvdata['polar'].append(float(data[resSeq]['polar'])/total_num_frames)
        csvdata['waterbridge'].append(float(data[resSeq]['waterbridge'])/total_num_frames)

    df = pd.DataFrame(csvdata)
    df.to_csv(args.out + '.csv', index=False, float_format='%.4f')
    g = df.loc[:, ~df.columns.isin(['resSeq','resName'])].plot.bar(
        x="resid", 
        stacked=True,
        title="Protein-Ligand Interactions", 
        xlabel='Residue', 
        ylabel='Fraction of MD trajectory', 
        figsize=(8,3), 
        fontsize=8,
        )
    fig = g.get_figure()
    fig.savefig(args.out + '.pdf', bbox_inches="tight", pad_inches=0.2, dpi=150)